# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from . import mercurial_repository_pb2 as mercurial__repository__pb2


class MercurialRepositoryServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.InitConfig = channel.unary_unary(
                '/hgitaly.MercurialRepositoryService/InitConfig',
                request_serializer=mercurial__repository__pb2.InitConfigRequest.SerializeToString,
                response_deserializer=mercurial__repository__pb2.InitConfigResponse.FromString,
                )
        self.GetConfigItem = channel.unary_unary(
                '/hgitaly.MercurialRepositoryService/GetConfigItem',
                request_serializer=mercurial__repository__pb2.GetConfigItemRequest.SerializeToString,
                response_deserializer=mercurial__repository__pb2.GetConfigItemResponse.FromString,
                )
        self.Recover = channel.unary_unary(
                '/hgitaly.MercurialRepositoryService/Recover',
                request_serializer=mercurial__repository__pb2.RecoverRequest.SerializeToString,
                response_deserializer=mercurial__repository__pb2.RecoverResponse.FromString,
                )
        self.Optimize = channel.unary_unary(
                '/hgitaly.MercurialRepositoryService/Optimize',
                request_serializer=mercurial__repository__pb2.OptimizeRequest.SerializeToString,
                response_deserializer=mercurial__repository__pb2.OptimizeResponse.FromString,
                )
        self.ResetCaches = channel.unary_unary(
                '/hgitaly.MercurialRepositoryService/ResetCaches',
                request_serializer=mercurial__repository__pb2.ResetCachesRequest.SerializeToString,
                response_deserializer=mercurial__repository__pb2.ResetCachesResponse.FromString,
                )
        self.GetManagedConfig = channel.unary_unary(
                '/hgitaly.MercurialRepositoryService/GetManagedConfig',
                request_serializer=mercurial__repository__pb2.GetManagedConfigRequest.SerializeToString,
                response_deserializer=mercurial__repository__pb2.GetManagedConfigResponse.FromString,
                )
        self.SetManagedConfig = channel.unary_unary(
                '/hgitaly.MercurialRepositoryService/SetManagedConfig',
                request_serializer=mercurial__repository__pb2.SetManagedConfigRequest.SerializeToString,
                response_deserializer=mercurial__repository__pb2.SetManagedConfigResponse.FromString,
                )
        self.Push = channel.unary_unary(
                '/hgitaly.MercurialRepositoryService/Push',
                request_serializer=mercurial__repository__pb2.PushRequest.SerializeToString,
                response_deserializer=mercurial__repository__pb2.PushResponse.FromString,
                )
        self.Pull = channel.unary_unary(
                '/hgitaly.MercurialRepositoryService/Pull',
                request_serializer=mercurial__repository__pb2.PullRequest.SerializeToString,
                response_deserializer=mercurial__repository__pb2.PullResponse.FromString,
                )


class MercurialRepositoryServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def InitConfig(self, request, context):
        """Will back Rails method `hg_create_repository`,
        currently implemented on `Gitlab::Mercurial::HgGitRepository`
        Together with `RepositoryService.CreateRepository` (see hgitaly!64)
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetConfigItem(self, request, context):
        """Will back Rails method `hg_config_item_bool?`  currently
        implemented on `Gitlab::Mercurial::HgGitRepository`
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Recover(self, request, context):
        """/ Recover after abandoned transaction
        /
        / Equivalent of `hg recover`
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Optimize(self, request, context):
        """/ Optimize repository
        /
        / Recompute stuff, switch to more efficient caches and compression, etc.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ResetCaches(self, request, context):
        """/ Reset caches utility
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetManagedConfig(self, request, context):
        """
        Managed configuration methods.


        These are about the part of the repository-local configuration that is
        entirely managed by Heptapod
        (implemented as of this writing with the separate `.hg/hgrc.managed` file)

        Will back Rails methods `get_hgrc` and `get_hg_heptapod_config` currently
        implemented on `Gitlab::Mercurial::HgGitRepository`
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SetManagedConfig(self, request, context):
        """Will back Rails methods `set_hgrc` currently implemented on
        `Gitlab::Mercurial::HgGitRepository` (but will be more restrictive,
        handling a closed list of sections and items, which is currently
        implemented by the upper layers of the Rails application).
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Push(self, request, context):
        """/ Will back Rails method `hg_push` (has never been implemented with a
        / subprocess)
        /
        / Main use-case is push mirroring, as in upstream's
        / `Remote.UpdateRemoteMirror`. A difference with the latter, that we can
        / accept for now, is the streaming request of `UpdateRemoteMirror`,
        / which is meant to pass very long lists of branch patterns in
        / `only_branches_matching`.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Pull(self, request, context):
        """/ Pull from some remote repo.
        /
        / Not very different from RepositoryService.FetchRemote, but we'll avoid
        / a few Git-specific concepts.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_MercurialRepositoryServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'InitConfig': grpc.unary_unary_rpc_method_handler(
                    servicer.InitConfig,
                    request_deserializer=mercurial__repository__pb2.InitConfigRequest.FromString,
                    response_serializer=mercurial__repository__pb2.InitConfigResponse.SerializeToString,
            ),
            'GetConfigItem': grpc.unary_unary_rpc_method_handler(
                    servicer.GetConfigItem,
                    request_deserializer=mercurial__repository__pb2.GetConfigItemRequest.FromString,
                    response_serializer=mercurial__repository__pb2.GetConfigItemResponse.SerializeToString,
            ),
            'Recover': grpc.unary_unary_rpc_method_handler(
                    servicer.Recover,
                    request_deserializer=mercurial__repository__pb2.RecoverRequest.FromString,
                    response_serializer=mercurial__repository__pb2.RecoverResponse.SerializeToString,
            ),
            'Optimize': grpc.unary_unary_rpc_method_handler(
                    servicer.Optimize,
                    request_deserializer=mercurial__repository__pb2.OptimizeRequest.FromString,
                    response_serializer=mercurial__repository__pb2.OptimizeResponse.SerializeToString,
            ),
            'ResetCaches': grpc.unary_unary_rpc_method_handler(
                    servicer.ResetCaches,
                    request_deserializer=mercurial__repository__pb2.ResetCachesRequest.FromString,
                    response_serializer=mercurial__repository__pb2.ResetCachesResponse.SerializeToString,
            ),
            'GetManagedConfig': grpc.unary_unary_rpc_method_handler(
                    servicer.GetManagedConfig,
                    request_deserializer=mercurial__repository__pb2.GetManagedConfigRequest.FromString,
                    response_serializer=mercurial__repository__pb2.GetManagedConfigResponse.SerializeToString,
            ),
            'SetManagedConfig': grpc.unary_unary_rpc_method_handler(
                    servicer.SetManagedConfig,
                    request_deserializer=mercurial__repository__pb2.SetManagedConfigRequest.FromString,
                    response_serializer=mercurial__repository__pb2.SetManagedConfigResponse.SerializeToString,
            ),
            'Push': grpc.unary_unary_rpc_method_handler(
                    servicer.Push,
                    request_deserializer=mercurial__repository__pb2.PushRequest.FromString,
                    response_serializer=mercurial__repository__pb2.PushResponse.SerializeToString,
            ),
            'Pull': grpc.unary_unary_rpc_method_handler(
                    servicer.Pull,
                    request_deserializer=mercurial__repository__pb2.PullRequest.FromString,
                    response_serializer=mercurial__repository__pb2.PullResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'hgitaly.MercurialRepositoryService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class MercurialRepositoryService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def InitConfig(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/hgitaly.MercurialRepositoryService/InitConfig',
            mercurial__repository__pb2.InitConfigRequest.SerializeToString,
            mercurial__repository__pb2.InitConfigResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetConfigItem(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/hgitaly.MercurialRepositoryService/GetConfigItem',
            mercurial__repository__pb2.GetConfigItemRequest.SerializeToString,
            mercurial__repository__pb2.GetConfigItemResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Recover(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/hgitaly.MercurialRepositoryService/Recover',
            mercurial__repository__pb2.RecoverRequest.SerializeToString,
            mercurial__repository__pb2.RecoverResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Optimize(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/hgitaly.MercurialRepositoryService/Optimize',
            mercurial__repository__pb2.OptimizeRequest.SerializeToString,
            mercurial__repository__pb2.OptimizeResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ResetCaches(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/hgitaly.MercurialRepositoryService/ResetCaches',
            mercurial__repository__pb2.ResetCachesRequest.SerializeToString,
            mercurial__repository__pb2.ResetCachesResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetManagedConfig(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/hgitaly.MercurialRepositoryService/GetManagedConfig',
            mercurial__repository__pb2.GetManagedConfigRequest.SerializeToString,
            mercurial__repository__pb2.GetManagedConfigResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def SetManagedConfig(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/hgitaly.MercurialRepositoryService/SetManagedConfig',
            mercurial__repository__pb2.SetManagedConfigRequest.SerializeToString,
            mercurial__repository__pb2.SetManagedConfigResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Push(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/hgitaly.MercurialRepositoryService/Push',
            mercurial__repository__pb2.PushRequest.SerializeToString,
            mercurial__repository__pb2.PushResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Pull(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/hgitaly.MercurialRepositoryService/Pull',
            mercurial__repository__pb2.PullRequest.SerializeToString,
            mercurial__repository__pb2.PullResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



# Copyright 2022 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
import grpc
import pytest

from hgitaly.stub.mercurial_changeset_pb2 import (
    ListMercurialChangesetsRequest,
    MercurialChangeset,
    MercurialChangesetField,
    MercurialChangesetPhase,
    MercurialRepositoryView,
)
from hgitaly.stub.mercurial_changeset_pb2_grpc import (
    MercurialChangesetServiceStub,
)

from .fixture import ServiceFixture


class ChangesetFixture(ServiceFixture):

    stub_cls = MercurialChangesetServiceStub

    def list_changesets(self, **kw):
        return [changeset
                for resp in self.stub.ListMercurialChangesets(
                        ListMercurialChangesetsRequest(
                            repository=self.grpc_repo, **kw)
                        )
                for changeset in resp.changesets
                ]


@pytest.fixture
def changeset_fixture(grpc_channel, server_repos_root):
    with ChangesetFixture(grpc_channel, server_repos_root) as fixture:
        yield fixture


def test_list_changesets(changeset_fixture):
    commit_file = changeset_fixture.repo_wrapper.commit_file
    default_ctx = commit_file('foo')
    default_hex = default_ctx.hex()
    changeset_fixture.repo_wrapper.set_phase('public', [default_hex])

    obs_hex = commit_file('foo', parent=default_ctx, topic='zetop').hex()
    topic_hex = changeset_fixture.repo_wrapper.amend_file('foo').hex()

    list_changesets = changeset_fixture.list_changesets

    ALL = MercurialChangesetField.ALL
    PHASE = MercurialChangesetField.PHASE
    DRAFT = MercurialChangesetPhase.DRAFT
    PUBLIC = MercurialChangesetPhase.PUBLIC
    OBSOLETE = MercurialChangesetField.OBSOLETE
    UNFILTERED = MercurialRepositoryView.UNFILTERED

    assert list_changesets(
        revset=obs_hex, fields=[OBSOLETE], view=UNFILTERED
    ) == [
        MercurialChangeset(id=obs_hex, obsolete=True)
    ]
    assert list_changesets(
        revset=obs_hex, fields=[ALL], view=UNFILTERED
    ) == [
        MercurialChangeset(id=obs_hex,
                           phase=DRAFT,
                           parent_ids=[default_hex],
                           branch=b'default',
                           topic=b'zetop',
                           topic_namespace=b'none',
                           fqbn=b'default//zetop',
                           obsolete=True,
                           )
    ]
    assert list_changesets(revset=b'::zetop', fields=[PHASE, OBSOLETE]) == [
        MercurialChangeset(id=default_hex, obsolete=False, phase=PUBLIC),
        MercurialChangeset(id=topic_hex, obsolete=False, phase=DRAFT),
    ]

    with pytest.raises(grpc.RpcError) as exc_info:
        list_changesets(revset=obs_hex, fields=[ALL])

    assert exc_info.value.code() == grpc.StatusCode.INVALID_ARGUMENT



# Copyright 2020-2022 Georges Racinet <georges.racinet@octobus.net>
# Copyright 2021 Sushil Khanchi <sushilkhanchi97@gmail.com>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
import itertools
import logging
import os

from grpc import StatusCode
from google.protobuf.timestamp_pb2 import Timestamp
from mercurial import (
    error,
    pycompat,
    logcmdutil,
    hgweb,
    node as nodemod,
)
from mercurial.node import nullrev as NULL_REVISION

from hgext3rd.heptapod.branch import get_default_gitlab_branch

from .. import (
    manifest,
    message,
)
from ..errors import (
    not_implemented,
    structured_abort,
)
from ..file_context import (
    git_perms,
)
from ..git import (
    OBJECT_MODE_TREE,
    GitPathSpec,
    GitLiteralPathSpec,
)
from ..linguist import (
    language_color,
    language_stats,
)
from ..logging import LoggerAdapter
from ..oid import (
    tree_oid,
    blob_oid,
)
from ..pagination import (
    extract_limit,
)
from ..revision import (
    RevisionNotFound,
    gitlab_revision_changeset,
    resolve_revspecs_positive_negative,
)
from ..revset import (
    FollowNotImplemented,
    changeset_descr_regexp,
    revset_from_git_revspec,
)
from ..servicer import HGitalyServicer
from ..stream import (
    concat_resplit,
    WRITE_BUFFER_SIZE,
)
from ..stub.commit_pb2 import (
    CommitIsAncestorRequest,
    CommitIsAncestorResponse,
    TreeEntryRequest,
    TreeEntryResponse,
    CheckObjectsExistRequest,
    CheckObjectsExistResponse,
    CountCommitsRequest,
    CountCommitsResponse,
    CountDivergingCommitsRequest,
    CountDivergingCommitsResponse,
    GetTreeEntriesError,
    GetTreeEntriesRequest,
    GetTreeEntriesResponse,
    TreeEntry,
    ListFilesRequest,
    ListFilesResponse,
    FindCommitRequest,
    FindCommitResponse,
    CommitStatsRequest,
    CommitStatsResponse,
    FindAllCommitsRequest,
    FindAllCommitsResponse,
    FindCommitsError,
    FindCommitsRequest,
    FindCommitsResponse,
    CommitLanguagesRequest,
    CommitLanguagesResponse,
    RawBlameError,
    RawBlameRequest,
    RawBlameResponse,
    LastCommitForPathRequest,
    LastCommitForPathResponse,
    ListLastCommitsForTreeRequest,
    ListLastCommitsForTreeResponse,
    CommitsByMessageRequest,
    CommitsByMessageResponse,
    ListCommitsByOidRequest,
    ListCommitsByOidResponse,
    ListCommitsRequest,
    ListCommitsResponse,
    ListCommitsByRefNameRequest,
    ListCommitsByRefNameResponse,
    FilterShasWithSignaturesRequest,
    FilterShasWithSignaturesResponse,
    GetCommitSignaturesRequest,
    GetCommitSignaturesResponse,
    GetCommitMessagesRequest,
    GetCommitMessagesResponse,
)
from ..stub.errors_pb2 import (
    PathError,
    PathNotFoundError,
    ResolveRevisionError,
)
from ..stub.commit_pb2_grpc import CommitServiceServicer
from ..util import (
    chunked,
    chunked_with_cursor,
)
from ..workdir import WorkingDirectoryError

base_logger = logging.getLogger(__name__)

NULL_REV = nodemod.nullrev
PSEUDO_REVS = (nodemod.wdirrev, nodemod.nullrev)
PSEUDO_REVLOG_NODES = {
    nodemod.nullid,
    nodemod.wdirid,
}
PSEUDO_REVLOG_NODES.update(nodemod.wdirfilenodeids)


class CommitServicer(CommitServiceServicer, HGitalyServicer):

    STATUS_CODE_STORAGE_NOT_FOUND = StatusCode.INVALID_ARGUMENT

    def CommitIsAncestor(self,
                         request: CommitIsAncestorRequest,
                         context) -> CommitIsAncestorResponse:
        logger = LoggerAdapter(base_logger, context)
        # The question is legit for filtered changesets and that
        # happens in MR rebase scenarios, before the Rails app realizes
        # the MR has to be updated.
        repo = self.load_repo(request.repository, context).unfiltered()
        # TODO status.Errorf(codes.InvalidArgument, "Bad Request
        # (empty ancestor sha)") and same for child
        try:
            ancestor = repo[request.ancestor_id.encode()]
            child = repo[request.child_id.encode()]
        except error.RepoLookupError as exc:
            # Gitaly just returns False. This is probably an inconsistency
            # in the client, so let's log it to help.
            logger.warning(
                "CommitIsAncestor for child_id=%r, ancestor_id=%r, got %r",
                request.ancestor_id, request.child_id, exc)
            result = False
        else:
            result = ancestor.isancestorof(child)

        return CommitIsAncestorResponse(value=result)

    def TreeEntry(self, request: TreeEntryRequest,
                  context) -> TreeEntryResponse:
        """Return an entry of a tree.

        The name could be confusing with the entry for a tree: the entry
        can be of any type.

        Actually, it always yields one response message, using the empty
        response in case the given path does not resolve.
        """
        repo = self.load_repo(request.repository, context).unfiltered()
        changeset = gitlab_revision_changeset(repo, request.revision)
        if changeset is None:
            # As of v15.8.0, Gitaly error details don't give up anything
            context.abort(StatusCode.NOT_FOUND, "tree entry not found")

        sha = changeset.hex().decode('ascii')
        # early testing shows that even for leaf files, Gitaly ignores
        # trailing slashes
        path = request.path.rstrip(b'/')

        try:
            filectx = changeset.filectx(path)
        except error.ManifestLookupError:
            filectx = None

        # TODO investigate why it's not the usual WRITE_BUFFER_SIZE
        # The only occurrence we could find so far was a 16384 in grpc Golang
        # lib (size of HTTP/2 frames).
        # Could be because Gitaly implementation uses CopyN to send to
        # its chunker (in `streamio.go`) and CopyN has a buffer.
        buffer_size = 16384
        if filectx is not None:
            otype = TreeEntryResponse.ObjectType.BLOB
            oid = blob_oid(repo, sha, path)
            mode = git_perms(filectx)
            size = filectx.size()
            max_size = request.max_size
            if max_size != 0 and size > request.max_size:
                context.abort(
                    StatusCode.FAILED_PRECONDITION,
                    "object size (%d) is bigger than the maximum "
                    "allowed size (%d)" % (size, max_size))

            data = filectx.data()

            limit = request.limit
            if limit != 0:
                data = data[:limit]

            offset = 0
            while offset < size:
                # only the first response of the stream carries the metadata
                if offset:
                    resp = TreeEntryResponse()
                else:
                    resp = TreeEntryResponse(type=otype,
                                             oid=oid,
                                             size=size,
                                             mode=mode)

                resp.data = data[offset:offset+buffer_size]
                offset += buffer_size

                yield resp
            return

        subtrees, file_paths = manifest.miner(changeset).ls_dir(path)
        if not subtrees and not file_paths:
            context.abort(StatusCode.NOT_FOUND, "tree entry not found")

        # path is an actual directory

        # size computation to match Git response.
        # The formula for size computation is meant to match the size
        # returned by Git, which is actually the size of the raw Git Tree
        # object as returned by `git-cat-file` with `<type>` parameter.
        #
        # The raw Git Tree object is a simple concatenation of entries, each
        # one being made of
        # - mode (octal representation): 6 bytes for blobs (e.g 100644),
        #                                5 bytes for subtrees (40000)
        # - 0x20 (separator): 1 byte
        # - name of the entry
        # - 0x00 (separator): 1 byte
        # - binary SHA-1 of the object referenced by the entry: 20 bytes
        # Hence the total length per entry is 27 + (name length) for subtrees
        # and 28 + (name length) for blobs.
        # Finally, our `ls_dir` returns full paths from the root, so the
        # have to substract `len(path + b'/')`, hence `len(path) + 1`
        # for each entry.
        size = (
            28 * len(file_paths) + 27 * len(subtrees)
            + sum(len(s) for s in subtrees) + sum(len(f) for f in file_paths)
            - (len(subtrees) + len(file_paths)) * (len(path) + 1)
        )
        # max_size does not apply here (see Gitaly comparison test)
        yield TreeEntryResponse(
            type=TreeEntryResponse.ObjectType.TREE,
            oid=tree_oid(repo, sha, path),
            mode=OBJECT_MODE_TREE,
            size=size,
        )

    def CountCommits(self,
                     request: CountCommitsRequest,
                     context) -> CountCommitsResponse:
        logger = LoggerAdapter(base_logger, context)
        # TODO: yet to finish this method to support all lookups
        repo = self.load_repo(request.repository, context)
        revision = request.revision
        # revision can be a pseudo range, like b'12340f9b5..a5f36b6a53012',
        # (see CommitsBetween for how we handle that)
        # (used in MR widget)
        if revision:
            if b'..' in revision:
                # TODO also case of ... (3 dots), I suppose
                ctx_start, ctx_end = [gitlab_revision_changeset(repo, rev)
                                      for rev in revision.split(b'..')]
                if ctx_start is None or ctx_end is None:
                    logger.warning(
                        "CountCommits for %r: one of these revisions "
                        "could not be found", revision)
                    return CountCommitsResponse()

                if ctx_end.obsolete() or ctx_start.obsolete():
                    repo = repo.unfiltered()

                revs = repo.revs('only(%s, %s)',
                                 ctx_end.hex(), ctx_start.hex())
            else:
                ctx = gitlab_revision_changeset(repo, revision)
                if ctx is None:
                    logger.warning(
                        "CountCommits revision %r could not be found",
                        revision)
                    return CountCommitsResponse()
                revs = repo.unfiltered().revs('::%s', ctx)
            count = len(revs)
        elif not request.all:
            # note: `revision` and `all` are mutually exclusive
            context.abort(StatusCode.INVALID_ARGUMENT,
                          "empty Revision and false All")
        else:
            # Note: if revision is not passed, we return all revs for now.
            # TODO not really exact, should be non obsolete and ::keeparounds
            count = len(repo)
        max_count = request.max_count
        if max_count and count > max_count:
            # TODO better to limit the revsets before hand
            count = max_count
        return CountCommitsResponse(count=count)

    # CountDivergingCommits counts the diverging commits between from and to.
    # Important to note that when --max-count is applied, the counts are not
    # guaranteed to be accurate.

    def CountDivergingCommits(self,
                              request: CountDivergingCommitsRequest,
                              context) -> CountDivergingCommitsResponse:
        logger = LoggerAdapter(base_logger, context)
        repo = self.load_repo(request.repository, context)
        rev_from = gitlab_revision_changeset(repo, getattr(request, 'from'))
        rev_to = gitlab_revision_changeset(repo, getattr(request, 'to'))
        max_count = request.max_count
        if rev_from is None:
            logger.warning("cannot resolve 'from' revision in %r",
                           message.Logging(request))
        if rev_to is None:
            logger.warning("cannot resolve 'to' revision in %r",
                           message.Logging(request))
        if rev_from is None or rev_to is None:
            return CountDivergingCommitsResponse(left_count=0, right_count=0)
        left = rev_from.rev()
        right = rev_to.rev()

        # Switching to unfiltered repo view, only if really needed.
        # using the unfiltered repo should not change the result given that
        # the revisions are fully resolved, but let's not take the risk.
        if rev_from.obsolete() or rev_to.obsolete():
            repo = repo.unfiltered()

        branchpoint = repo.revs(b"ancestor(%d, %d)" % (left, right)).first()
        if branchpoint is None:
            left_revset = b'::%d' % left
            right_revset = b'::%d' % right
        else:
            left_revset = b"%d::%d - %d" % (branchpoint, left, branchpoint)
            right_revset = b"%d::%d  - %d" % (branchpoint, right, branchpoint)
        left_count = len(repo.revs(left_revset))
        right_count = len(repo.revs(right_revset))

        if max_count and (left_count + right_count) > max_count:
            delta = (left_count + right_count) - max_count
            if left_count >= delta:
                left_count -= delta
            else:
                delta -= left_count
                left_count = 0
                right_count -= delta
        return CountDivergingCommitsResponse(left_count=left_count,
                                             right_count=right_count)

    def GetTreeEntries(self, request: GetTreeEntriesRequest,
                       context) -> GetTreeEntriesResponse:
        repo = self.load_repo(request.repository, context)
        revision = request.revision
        changeset = gitlab_revision_changeset(repo, revision)
        if changeset is None:
            structured_abort(
                context,
                StatusCode.INVALID_ARGUMENT,
                "invalid revision or path",
                GetTreeEntriesError(resolve_tree=ResolveRevisionError(
                    revision=request.revision)))

        repo = repo.unfiltered()
        sha = changeset.hex().decode('ascii')
        if not request.path:
            structured_abort(
                context,
                StatusCode.INVALID_ARGUMENT, "empty path",
                GetTreeEntriesError(
                    path=PathError(path=request.path,
                                   error_type=PathError.ERROR_TYPE_EMPTY_PATH))
            )
        path = request.path.rstrip(b'/')  # same as in TreeEntry
        if path == b'.':  # special case, means the top directory
            path = b''

        blob_type = TreeEntry.EntryType.BLOB
        tree_type = TreeEntry.EntryType.TREE
        miner = manifest.miner(changeset)

        if request.recursive:
            entries = ((epath, is_dir, b'')
                       for epath, is_dir in miner.iter_dir_recursive(path))
        elif request.skip_flat_paths:
            trees, files = miner.ls_dir(path)
            entries = itertools.chain(
                ((epath, True, b'') for epath in trees),
                ((epath, False, b'') for epath in files),
            )
        else:
            entries = miner.iter_dir_with_flat_paths(path)

        # TODO request.sort == SortBy.TREES_FIRST
        limit = extract_limit(request)
        if limit == 0:
            return  # simplify things (next_cursor computation notably)

        # voids the advantages of iteration, but there's few choice
        # in the matter with sort and pagination
        if request.sort == GetTreeEntriesRequest.SortBy.TREES_FIRST:
            # each entry is (path, is_dir (bool), flat_path)
            # where flat_path is ignored for sorting purposes.
            # TODO OPTIM is it an improvement to avoid this post-processing
            # sort by making ManifestMiner already provide directories first?
            # to get an advantage, we'd need to avoid at least a list
            # construction
            entries = sorted(
                entries,
                key=lambda entry: (1 if entry[1] else 2, entry[0]))
        else:
            # TODO OPTIM because Mercurial manifests are stored in
            # lexicographical ordering, this sort is probably unnecessary.
            # In first impl of sorting and pagination, probably not worth
            # the risk.
            # Also would be neat to avoid consuming the iterator into
            # a least, but is it worth the added complexity if it can't
            # be done in the TREES_FIRST case?
            # What if GitLab introduces a sort direction anyway ?
            entries = sorted(entries)  # sorts lexicographically by path

        if not entries:
            # there is no such thing as an empty directory in Mercurial
            try:
                changeset.filectx(path)
            except error.ManifestLookupError:
                if request.recursive:
                    err_code = StatusCode.NOT_FOUND
                else:
                    err_code = StatusCode.INVALID_ARGUMENT
                structured_abort(
                    context,
                    err_code,
                    "invalid revision or path",
                    GetTreeEntriesError(resolve_tree=ResolveRevisionError(
                        revision=request.revision))
                )
            else:
                structured_abort(
                    context,
                    StatusCode.INVALID_ARGUMENT,
                    "path not treeish",
                    GetTreeEntriesError(resolve_tree=ResolveRevisionError(
                        revision=request.revision))
                )

        def entry_oid(path, is_dir):
            return (tree_oid if is_dir else blob_oid)(repo, sha, path)

        page_token = request.pagination_params.page_token
        if page_token:
            for offset, (path, is_dir, _) in enumerate(entries):
                if entry_oid(path, is_dir) == page_token:
                    break
            else:
                context.abort(StatusCode.INTERNAL,
                              "could not find starting OID: %s" % page_token)

            entries = entries[offset+1:]

        if limit < len(entries):
            entries = entries[:limit]
            last_path, last_is_dir = entries[-1][:2]
            next_cursor = entry_oid(last_path, last_is_dir)
        else:
            next_cursor = ''

        yield from chunked_with_cursor(
            GetTreeEntriesResponse,
            entries,
            next_cursor=next_cursor,
            builder=lambda chunk: dict(
                entries=(
                    TreeEntry(
                        path=path,
                        type=tree_type if is_dir else blob_type,
                        oid=entry_oid(path, is_dir),
                        commit_oid=revision,
                        mode=(OBJECT_MODE_TREE if is_dir else
                              git_perms(changeset.filectx(path))),
                        flat_path=flat_path,
                    )
                    for path, is_dir, flat_path in chunk)
                )
        )

    def ListFiles(self, request: ListFilesRequest,
                  context) -> ListFilesResponse:
        repo = self.load_repo(request.repository, context)
        revision = pycompat.sysbytes(request.revision)
        ctx = gitlab_revision_changeset(repo, revision)
        if ctx is None:
            return
        mf = ctx.manifest()
        for paths in chunked(mf.iterkeys()):
            yield ListFilesResponse(paths=paths)

    def CommitStats(self, request: CommitStatsRequest,
                    context) -> CommitStatsResponse:
        repo = self.load_repo(request.repository, context)
        revision = pycompat.sysbytes(request.revision)
        ctx = gitlab_revision_changeset(repo, revision)
        if ctx is None:
            context.abort(StatusCode.INTERNAL,
                          "failed to get commit stats: object not found.")

        ctxp1 = ctx.p1()
        statsgen = hgweb.webutil.diffstatgen(repo.ui, ctx, ctxp1)
        # stats format:
        #   (list_of_stats_per_file, maxname,
        #    maxtotal, addtotal, removetotal, binary)
        # we only need addtotal and removetotal for our use case
        stats = next(statsgen)
        addtotal, removetotal = stats[-3], stats[-2]
        return CommitStatsResponse(
            oid=ctx.hex(),
            additions=addtotal,
            deletions=removetotal,
        )

    def FindCommit(self,
                   request: FindCommitRequest, context) -> FindCommitResponse:
        logger = LoggerAdapter(base_logger, context)
        revision = request.revision
        if not revision:
            context.abort(StatusCode.INVALID_ARGUMENT, "empty revision")

        repo = self.load_repo(request.repository, context)
        ctx = gitlab_revision_changeset(repo, revision)

        if ctx is None:
            logger.warning("FindCommit revision %r could not be found",
                           revision)
            return FindCommitResponse()

        if ctx.rev() == NULL_REV:
            return FindCommitResponse()

        commit = message.commit(ctx)
        return FindCommitResponse(commit=commit)

    def FindAllCommits(self, request: FindAllCommitsRequest,
                       context) -> FindAllCommitsResponse:
        logger = LoggerAdapter(base_logger, context)
        repo = self.load_repo(request.repository, context)
        revision = request.revision
        if revision:
            # If false, return all commits reachable by any branch in the repo
            ctx = gitlab_revision_changeset(repo, revision)
            if ctx is None:
                logger.warning(
                    "FindAllCommits revision %r could not be found",
                    revision)
                return FindAllCommitsResponse()
            revset = b"reverse(::%s)" % ctx
            # if ctx is an obsolete changeset, its repo is unfiltered.
            # this is legitimate if revision is a direct hash or a GitLab
            # special ref and should not happen otherwise
            repo = ctx.repo()
        else:
            revset = b'reverse(all())'

        revs = repo.revs(revset)
        offset = request.skip
        if offset and offset > 0:
            revs = revs.slice(offset, len(revs))
        if request.max_count:
            revs = revs.slice(0, request.max_count)
        if request.order == FindAllCommitsRequest.TOPO:
            revs = repo.revs(b"sort(%ld, topo)", revs)
        elif request.order == FindAllCommitsRequest.DATE:
            revs = repo.revs(b"reverse(sort(%ld, date))", revs)
        for chunk in chunked(revs):
            yield FindAllCommitsResponse(
                commits=(message.commit(repo[rev]) for rev in chunk))

    def FindCommits(self, request: FindCommitsRequest,
                    context) -> FindCommitsResponse:

        logger = LoggerAdapter(base_logger, context)
        req_log = message.Logging(request)

        if request.limit == 0:
            return

        repo = self.load_repo(request.repository, context)
        pats = request.paths
        # XXX: raise error if one of the path given is an empty string
        if pats:
            pats = list(map(lambda p: repo.root + b'/' + p, pats))

        repo, opts = parse_find_commits_request_opts(request, context, repo)

        if request.revision and not opts[b'rev'][0]:
            logger.debug(
                "Request %r, revision could not be found", req_log)
            structured_abort(
                context, StatusCode.NOT_FOUND, "commits not found",
                FindCommitsError())

        walk_opts = logcmdutil.parseopts(repo.ui, pats, opts)
        revs, _ = logcmdutil.getrevs(repo, walk_opts)

        if request.offset > 0:
            revs = revs.slice(request.offset, len(revs))

        if request.order == FindCommitsRequest.TOPO:
            revs = repo.revs(b"sort(%ld, topo)", revs)
            if request.all:
                revs = repo.revs(b"reverse(%ld)", revs)

        # investigation log for heptapod#1365
        if len(revs) == 0:
            logger.info("Empty commit list for walk_opts=%r", walk_opts)
#            context.abort(StatusCode.NOT_FOUND, "commits not found")

        incl_ref_by = request.include_referenced_by

        with_short_stats = request.include_shortstat
        for chunk in chunked(revs):
            yield FindCommitsResponse(
                commits=(message.commit(repo[rev],
                                        include_referenced_by=incl_ref_by,
                                        with_short_stats=with_short_stats)
                         for rev in chunk))

    def CommitLanguages(self, request: CommitLanguagesRequest,
                        context) -> CommitLanguagesResponse:
        logger = LoggerAdapter(base_logger, context)
        repo = self.load_repo(request.repository, context)
        changeset = gitlab_revision_changeset(repo, request.revision)
        tokei_bin_path = os.fsdecode(
            repo.ui.config(b'hgitaly', b'tokei-executable')
        )
        try:
            with self.working_dir(gl_repo=request.repository,
                                  repo=repo,
                                  context=context,
                                  changeset=changeset) as wd:

                stats, total = language_stats(tokei_bin_path, wd.path, context)
                Language = CommitLanguagesResponse.Language
                resp_langs = []
                for lang_name, lang_stats in stats.items():
                    resp_langs.append(Language(
                        name=lang_name,
                        share=100 * lang_stats['total_bytes'] / total,
                        bytes=lang_stats['total_bytes'],
                        color=language_color(lang_name),
                    ))
                    resp_langs.sort(key=lambda lang: -lang.share)
                return CommitLanguagesResponse(languages=resp_langs)
        except WorkingDirectoryError as exc:
            logger.warning("Failed to acquire a working directory"
                           "at stage %r" % exc.args[0])
            return CommitLanguagesResponse()

    def RawBlame(self, request: RawBlameRequest,
                 context) -> RawBlameResponse:
        repo = self.load_repo(request.repository, context)
        filepath = request.path
        if not filepath:
            context.abort(StatusCode.INVALID_ARGUMENT, "empty Path")

        if request.range:
            lstart, lend = [int(x) for x in request.range.split(b',')]
            lstart -= 1
        else:
            lstart, lend = 0, None

        revision = pycompat.sysbytes(request.revision)
        ctx = gitlab_revision_changeset(repo, revision)
        if ctx is None:
            return
        try:
            for data in concat_resplit(blamelines(repo, ctx, filepath,
                                                  lstart=lstart,
                                                  lend=lend,
                                                  ),
                                       WRITE_BUFFER_SIZE):
                yield RawBlameResponse(data=data)
        except error.ManifestLookupError:
            structured_abort(
                context, StatusCode.NOT_FOUND,
                "path not found in revision",
                RawBlameError(path_not_found=PathNotFoundError(path=filepath))
            )
        except BlameRangeError as exc:
            structured_abort(
                context,
                StatusCode.INVALID_ARGUMENT,
                "range is outside of the file length",
                RawBlameError(out_of_range=RawBlameError.OutOfRangeError(
                    actual_lines=exc.actual_lines())))

    def LastCommitForPath(self,
                          request: LastCommitForPathRequest,
                          context) -> LastCommitForPathResponse:
        repo = self.load_repo(request.repository, context)
        revision, path = request.revision, request.path
        ctx = gitlab_revision_changeset(repo, revision)
        if ctx is None or ctx.rev() == NULL_REVISION:
            context.abort(StatusCode.INTERNAL,
                          "logging last commit for path: exit status 128")

        changeset = latest_changeset_for_path(
            path, ctx,
            literal_pathspec=(request.global_options.literal_pathspecs
                              or request.literal_pathspec)
        )

        return LastCommitForPathResponse(
            commit=message.commit(changeset),
        )

    def ListLastCommitsForTree(self, request: ListLastCommitsForTreeRequest,
                               context) -> ListLastCommitsForTreeResponse:
        repo = self.load_repo(request.repository, context)
        revision = pycompat.sysbytes(request.revision)
        from_ctx = gitlab_revision_changeset(repo, revision)
        if from_ctx is None or from_ctx.rev() == NULL_REVISION:
            context.abort(StatusCode.INTERNAL, "exit status 128")

        offset, limit = request.offset, request.limit
        if limit == 0:
            return

        if limit < 0:
            context.abort(StatusCode.INVALID_ARGUMENT, 'limit negative')
        if offset < 0:
            context.abort(StatusCode.INVALID_ARGUMENT, 'offset negative')

        req_path = request.path

        if req_path in (b'.', b'/', b'./'):
            req_path = b''

        if req_path and not req_path.endswith(b'/'):
            if offset > 0:
                return

            changeset = latest_changeset_for_path(req_path, from_ctx,
                                                  literal_pathspec=True)
            if changeset:
                yield ListLastCommitsForTreeResponse(
                    commits=[message.commit_for_tree(changeset, req_path)])
            return

        # subtrees first, then regular files, each one in lexicographical order
        subtrees, file_paths = manifest.miner(from_ctx).ls_dir(req_path)
        end = offset + limit
        nb_subtrees = len(subtrees)
        if nb_subtrees > end:
            subtrees = subtrees[offset:end]
            file_paths = ()
        else:
            subtrees = subtrees[offset:end]
            file_paths = file_paths[max(offset-nb_subtrees, 0):end-nb_subtrees]

        changesets = latest_changesets_for_paths(subtrees, file_paths,
                                                 from_ctx)
        all_paths = subtrees
        all_paths.extend(file_paths)
        for chunk in chunked(all_paths):
            yield ListLastCommitsForTreeResponse(
                commits=[
                    message.commit_for_tree(
                        changesets[path],
                        path
                    )
                    for path in chunk
                ])

    def CommitsByMessage(self, request: CommitsByMessageRequest,
                         context) -> CommitsByMessageResponse:
        logger = LoggerAdapter(base_logger, context)
        repo = self.load_repo(request.repository, context)
        query = request.query
        if not query:
            return CommitsByMessageResponse()
        pats = []
        opts = {}
        if request.path:
            path = repo.root + b'/' + request.path
            pats.append(path)
        if request.limit:
            opts[b'limit'] = request.limit
        if request.revision:
            revset = revset_from_git_revspec(repo, request.revision)
            repo = repo.unfiltered()
            if revset is None:
                logger.debug(
                    "CommitsByMessage revision %r could not be found",
                    request.revision)
                return CommitsByMessageResponse()
        else:
            revision = get_default_gitlab_branch(repo)
            # XXX: return error if no default branch found
            revset = revset_from_git_revspec(repo, revision)
        # Instead of sending 'query' as a key:value pair ('keyword': query) in
        # `opts`, appending the query to `revset` as "...and keyword('query')"
        # to make sure it perform an intersetion of two, instead of a union.
        revset = revset + b" and keyword('%b')" % query.encode()
        opts[b'rev'] = [revset]
        walk_opts = logcmdutil.parseopts(repo.ui, pats, opts)
        revs, _ = logcmdutil.getrevs(repo, walk_opts)
        offset = request.offset
        if offset and offset > 0:
            revs = revs.slice(offset, len(revs))
        for chunk in chunked(revs):
            yield CommitsByMessageResponse(
                commits=(message.commit(repo[rev]) for rev in chunk))

    def CheckObjectsExist(self, request: CheckObjectsExistRequest,
                          context) -> CheckObjectsExistResponse:
        not_implemented(context, issue=101)  # pragma no cover

    def ListCommits(self, request: ListCommitsRequest,
                    context) -> ListCommitsResponse:
        """Implementation of ListCommits, with some differences with Gitaly.

        Orderings
        ~~~~~~~~~
        As recalled in `commit.proto` comment, Git default ordering (and
        hence Gitaly's) is by date BUT that means actually first by
        parentship, then by date (actually CommitDate, not AuthorDate).
        Quoting git-log(1):

         --date-order
             Show no parents before all of its children are shown, but
             otherwise show commits in the commit timestamp order.

        I could check that this commit timestamp is the `CommitDate` field.

        On the other hand, Mercurial's date ordering ignores
        the graph completely, and there's no mixed ordering similar to Git's
        (although perhaps the `topo.firstbranch` could be used for this.

        NONE (default)
        --------------
        By default, instead of Git's (parentship, CommitDate) we're using
        the rev number: it respects parentship, and is conceptually close
        to the CommitDate ordering, which is by default the actual date of
        this exact commit creation (for instance, it is updated by
        `git commit --amend` whereas AuthodDate is not). CommitDate can be
        forced on a Git repository, but there's no Mercurial equivalent of
        that. The end result will be something akin to a PushDate field if
        there was any.

        There is hopefully no logic that really depends on the ordering on
        the client side, as long as it respects the parentship. As of this
        writing, this call is used for Merge Requests list of commits.
        Respecting the parentship is important, CommitDate on the Git side
        vs PushDate on the Mercurial side is probably not.

        TOPO
        ----

        These behave the same way between Git and Mercurial, the client
        will have to consider the choice of parent to present first to
        be arbitrary anyway

        DATE
        ----

        As explained above, Mercurial's date ordering is quite different
        from Git's. For now we choose to use it nevertheless if explicitely
        requested, but this could change if we identify trouble.
        """
        revisions = request.revisions
        if not revisions:
            context.abort(StatusCode.INVALID_ARGUMENT, "missing revisions")

        repo = self.load_repo(request.repository, context)
        revisions = (pycompat.sysbytes(r) for r in revisions)
        try:
            positive, negative = resolve_revspecs_positive_negative(
                repo, revisions)
        except RevisionNotFound as exc:
            context.abort(StatusCode.INTERNAL,
                          "Revision %r could not be resolved" % exc.args[0])

        # now that resolution is done, we can and have to switch to
        # the unfiltered view, because the resulting hashes can be of
        # obsolete changesets.
        repo = repo.unfiltered()
        if negative:
            revset = b'only(%ls, %ls)'
        else:
            revset = b'::%ls'

        msg_patterns = request.commit_message_patterns
        ignore_case = request.ignore_case
        if msg_patterns:
            # TODO all kinds of painful escapings (should be in lib)
            greps = [changeset_descr_regexp(pattern, ignore_case=ignore_case)
                     for pattern in msg_patterns]
            if len(greps) > 1:
                revset += b' and (%s)' % b' or '.join(greps)
            else:
                revset += b' and ' + greps[0]

        Order = ListCommitsRequest.Order
        reverse = request.reverse

        if request.order == Order.NONE:
            # default Git ordering is Mercurial's reversed
            sort_key = b'rev' if reverse else b'-rev'
        elif request.order == Order.TOPO:
            sort_key = b'-topo' if reverse else b'topo'
        elif request.order == Order.DATE:
            # See docstring about this approximative choice
            sort_key = b'-date' if reverse else b'date'

        revset = b'sort(%s, %s)' % (revset, sort_key)

        try:
            if negative:
                revs = repo.revs(revset, positive, negative)
            else:
                revs = repo.revs(revset, positive)
        except error.ParseError as exc:
            # no point trying to mimic Gitaly's error message, as it is
            # very dependent on internal details. Example for invalid regexp
            # with Gitaly 15.4:
            #  iterating objects: re...e command: exit status 128,
            #  stderr: "fatal: command line, '[]': Unmatched [, [^, [:, [.,
            #  or [=\n"
            context.abort(StatusCode.INTERNAL, "Invalid filter: " + str(exc))

        # according to protocol comment, the page token
        # is the last commit OID already sent (similar to blob/tree requests)
        cursor = request.pagination_params.page_token
        if not cursor:
            offset = 0
        else:
            # TODO perf it would probably be much faster to use the index
            # directly rather than to construct contexts
            for offset, rev in enumerate(revs):
                sha = repo[rev].hex().decode()
                if sha == cursor:
                    offset = offset + 1
                    break
        limit = extract_limit(request)
        revs = revs.slice(offset, offset + limit)
        if not revs:
            return

        for chunk in chunked(revs):
            yield ListCommitsResponse(
                commits=(message.commit(repo[rev]) for rev in chunk))

    def ListCommitsByOid(self, request: ListCommitsByOidRequest,
                         context) -> ListCommitsByOidResponse:
        repo = self.load_repo(request.repository, context)
        repo = repo.unfiltered()
        lookup_error_classes = (error.LookupError, error.RepoLookupError)
        for chunk in chunked(pycompat.sysbytes(oid) for oid in request.oid):
            try:
                chunk_commits = [message.commit(repo[rev])
                                 for rev in repo.revs(b'%ls', chunk)
                                 if rev not in PSEUDO_REVS]
            except lookup_error_classes:
                # lookup errors aren't surprising: the client uses this
                # method for prefix resolution
                # The reference Gitaly implementation tries them one after
                # the other (as of v13.4.6)
                chunk_commits = []
                for oid in chunk:
                    try:
                        # TODO here, something only involving the nodemap
                        # would be in order
                        revs = repo.revs(b'%s', oid)
                    except lookup_error_classes:
                        # ignore unresolvable oid prefix
                        pass
                    else:
                        if len(revs) == 1:
                            rev = revs.first()
                            if rev in PSEUDO_REVS:
                                continue
                            chunk_commits.append(message.commit(repo[rev]))
            if chunk_commits:
                yield ListCommitsByOidResponse(commits=chunk_commits)

    def ListCommitsByRefName(self, request: ListCommitsByRefNameRequest,
                             context) -> ListCommitsByRefNameResponse:
        logger = LoggerAdapter(base_logger, context)
        repo = self.load_repo(request.repository, context)
        ref_names = request.ref_names

        commits = []
        for ref_name in ref_names:
            ctx = gitlab_revision_changeset(repo, ref_name)
            if ctx is None:
                logger.warning(
                    "ListCommitByRefName ref %r could not be "
                    "resolved to a changeset",
                    ref_name)
                continue
            commits.append([ref_name, ctx])
        CommitForRef = ListCommitsByRefNameResponse.CommitForRef
        for chunk in chunked(commits):
            yield ListCommitsByRefNameResponse(
                commit_refs=(CommitForRef(
                    commit=message.commit(ctx),
                    ref_name=ref_name
                ) for ref_name, ctx in chunk)
            )

    def FilterShasWithSignatures(self,
                                 request: FilterShasWithSignaturesRequest,
                                 context) -> FilterShasWithSignaturesResponse:
        not_implemented(context, issue=24)  # pragma no cover

    def GetCommitSignatures(self, request: GetCommitSignaturesRequest,
                            context) -> GetCommitSignaturesResponse:
        not_implemented(context, issue=24)  # pragma no cover

    def GetCommitMessages(self, request: GetCommitMessagesRequest,
                          context) -> GetCommitMessagesResponse:
        repo = self.load_repo(request.repository, context)
        results = {}
        for commit_id in request.commit_ids:
            commit_id = pycompat.sysbytes(commit_id)
            ctx = gitlab_revision_changeset(repo, commit_id)
            if ctx is None:
                # should not be an "internal" error, but
                # that's what Gitaly does anyway
                context.abort(
                    StatusCode.INTERNAL,
                    "failed to get commit message: object not found.")

            results[commit_id] = ctx.description()
        for commit_id, msg in results.items():
            yield GetCommitMessagesResponse(commit_id=commit_id,
                                            message=msg)


def parse_find_commits_request_opts(request, context, repo):
    """Interpred FindCommitRequestAttributes

    :return: (repo, options for logcmdutil.parseopts). Returning the repo
      is important because we often (but not always) need to switch to
      the unfiltered repo.
    """
    logger = LoggerAdapter(base_logger, context)
    opts = {
        b'follow': request.follow,
        b'no_merges': request.skip_merges,
        b'limit': request.limit + request.offset,
    }
    # TODO: implement 'request.first_parent' option
    # khanchi97: found that its counterpart follow-first in "hg log" is
    # deprecated and give wrong results with other options like revision,
    # all, etc.
    if request.author:
        opts[b'user'] = [request.author]
    after = request.after.ToSeconds()
    before = request.before.ToSeconds()
    date = getdate(after, before)
    if date is not None:
        opts[b'date'] = date

    revision = request.revision
    # `revision` and `all` are mutually exclusive,
    # if both present `all` gets the precedence
    if request.all:
        opts[b'rev'] = [b'0:tip']
        return repo, opts

    if not revision:
        revision = get_default_gitlab_branch(repo)

    if revision:
        try:
            revset = revset_from_git_revspec(repo, revision,
                                             for_follow=request.follow)
        except FollowNotImplemented:
            logger.warning("Ignoring `follow: true` (not implemented with "
                           "this revspec) for %r", message.Logging(request))
            revset = revset_from_git_revspec(repo, revision)
            opts[b'follow'] = False

        opts[b'rev'] = [revset]
    return repo.unfiltered(), opts


def getdate(after, before):
    if after and before:
        after = _isoformat_from_seconds(after)
        before = _isoformat_from_seconds(before)
        return "%s UTC to %s UTC" % (after, before)
    elif after:
        after = _isoformat_from_seconds(after)
        return ">%s UTC" % after
    elif before:
        before = _isoformat_from_seconds(before)
        return "<%s UTC" % before
    return None


def _isoformat_from_seconds(secs):
    ts = Timestamp()
    ts.FromSeconds(int(secs))
    dt = ts.ToDatetime()
    return dt.isoformat()


def ancestors_inclusive(changeset):
    yield changeset
    yield from changeset.ancestors()


def latest_changeset_for_path(path, seen_from, literal_pathspec=False):
    """Return latest ancestor of ``seen_from`` that touched the given path.

    :param bytes path: subdir or file
    :param seen_from: changectx
    """
    if not path or path in (b'.', b'/', b'./'):
        # interpeted as a directory, this would be the repository root, which
        # all changesets change.
        # TODO comment above is wrong, as empty changesets do exist!
        # The question would be what Gitaly does in this case, if we care.
        return seen_from

    if literal_pathspec:
        match = GitLiteralPathSpec(path).match
    else:
        match = GitPathSpec(path).match

    for changeset in ancestors_inclusive(seen_from):
        if any(match(fp) for fp in changeset.files()):
            return changeset


def latest_changesets_for_paths(subdirs, file_paths, seen_from):
    """Return latest ancestor of ``seen_from`` that touched the given path.

    :param bytes path: subdir or file
    :param seen_from: changectx
    """
    dir_prefixes = set(d.rstrip(b'/') + b'/' for d in subdirs)
    file_paths = set(file_paths)

    latest_changesets = {}

    for changeset in ancestors_inclusive(seen_from):
        if not dir_prefixes and not file_paths:
            break

        for cfp in changeset.files():
            # TODO some performance improvement possible with a binary
            # search if changeset files are guaranteed to be sorted
            # but it'd probably make more of a difference in a Rust impl
            for dp in dir_prefixes:
                if cfp.startswith(dp):
                    latest_changesets[dp.rstrip(b'/')] = changeset
                    dir_prefixes.discard(dp)
                    break
            else:
                for fp in file_paths:
                    if fp == cfp:
                        latest_changesets[fp] = changeset
                        break
                else:
                    continue
                file_paths.discard(fp)
                continue

            dir_prefixes.discard(dp)

    return latest_changesets


class BlameRangeError(RuntimeError):
    def actual_lines(self):
        return self.args[0]


def blamelines(repo, ctx, file, lstart=0, lend=None):
    """Yield blame lines of a file.
    """
    fctx = ctx[file]
    # fctx.annotate() does not seem to be able to be linited to
    # a range, and is not even an iterator (as of Mercurial 6.6).
    # All we can do is use islice for the day it would become an iterator.
    # so that, e.g.,  annotating only the first line of a very large file
    # is less expensive than annotating the whole
    annotated = fctx.annotate()
    if lstart >= len(annotated):
        raise BlameRangeError(len(annotated))

    for line_no, line in enumerate(
            itertools.islice(fctx.annotate(), lstart, lend),
            start=1):
        old_line_no = line.lineno
        # required blame line format that get parsed by Rails:
        #   '<hash_id> <old_line_no> <line_no>\n\t<line_text>'
        yield b'%s %d %d\n\t%s' % (line.fctx.hex(), old_line_no, line_no,
                                   line.text)



# Copyright 2020-2023 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
from contextlib import contextmanager
from grpc import StatusCode
from google.protobuf.any_pb2 import Any
from google.rpc.status_pb2 import Status
import logging

from .stub.errors_pb2 import (
    CustomHookError,
)

from heptapod.gitlab.hooks import (
    GitLabPreReceiveError,
    GitLabPostReceiveError,
)

logger = logging.getLogger(__name__)

HGITALY_ISSUES_URL = "https://foss.heptapod.net/heptapod/hgitaly/-/issues"
STATUS_DETAILS_KEY = 'grpc-status-details-bin'


class ServiceError(RuntimeError):
    """An exception class to complement setting of context.

    In cases where a more precise exception than the bare `Exception()` raised
    by `ServicerContext.abort()` is useful.

    Caller is expected to set code and optionally details.
    """


def not_implemented(context, issue: int):
    """Raise with NOT_IMPLEMENTED status code and link to issue.
    """
    msg = "Not implemented. Tracking issue: %s/%d" % (HGITALY_ISSUES_URL,
                                                      issue)
    logger.error(msg)
    context.abort(StatusCode.UNIMPLEMENTED, msg)


def structured_abort(context, code, msg, structured_error):
    """Abort method with Gitaly's structured error.
    """
    metadata = context.trailing_metadata()
    # ensure mutability (as a list since that is how we'll do it)
    metadata = [] if metadata is None else list(metadata)

    as_grpc_any = Any()
    as_grpc_any.Pack(structured_error)
    status = Status(code=code.value[0], message=msg, details=[as_grpc_any])
    metadata.append((STATUS_DETAILS_KEY, status.SerializeToString()))
    context.set_trailing_metadata(metadata)
    context.abort(code, msg)


def msg_type_url(msg_cls):
    """Return the type url associated with `msg_cls` in `Any` messages."""
    sample = Any()
    sample.Pack(msg_cls())
    return sample.type_url


def parse_structured_error(exc, msg_cls):
    """Test helper for structured errors.

    Asserts that the structured error has the given Message class.

    :returns: (numeric code, user message, instance of `msg_cls`)
    """
    for md in exc.trailing_metadata():
        if md.key == STATUS_DETAILS_KEY:
            status = Status.FromString(md.value)
            break
    else:  # pragma no cover (case just to make tests easier to debug)
        raise LookupError("gRPC status details not found in trailing metadata")
    details = status.details[0]

    assert details.type_url == msg_type_url(msg_cls)

    return status.code, status.message, msg_cls.FromString(details.value)


def parse_assert_structured_error(exc, msg_cls, expected_code):
    """Assert code and type and return details and message object.
    """
    assert exc.code() == expected_code
    code, details, msg = parse_structured_error(exc, msg_cls)
    assert code == expected_code.value[0]
    return details, msg


@contextmanager
def operation_error_treatment(context, error_message_class, logger,
                              error_message_attr='gitlab_hook'):
    try:
        yield
    except GitLabPreReceiveError as exc:
        message = exc.message
        if isinstance(message, str):
            # should not normally happen, but let's make sure it is not
            # a problem
            message = message.encode('utf-8')
        attrs = {
            error_message_attr: CustomHookError(
                hook_type=CustomHookError.HOOK_TYPE_PRERECEIVE,
                stderr=message
            )
        }
        structured_abort(context, StatusCode.PERMISSION_DENIED,
                         "not allowed",
                         error_message_class(**attrs))
    except GitLabPostReceiveError as exc:
        # post-receive errors should not abort. Quoting current (v16.6)
        # Gitaly tests for UserMergeBranch:
        #   // The post-receive hook runs after references have been updated
        #   // and any failures of it are ignored.
        logger.error("Error in GitLab post-receive hook: %r", exc.message)


class MercurialNotFound(LookupError):
    """General class for all failed lookups.

    Better for us that to use `hg.errors.LookupError` because the latter is
    not systematically raised. Anyway, we need our own articulation point.
    """


class MercurialPathNotFound(MercurialNotFound):
    """Express that some path could not be found.

    This is generally in the context of some changeset that the caller is
    aware of, but it could be in an entire revset, in the workdir…
    """


class MercurialChangesetNotFound(MercurialNotFound):
    """Express that one or serveral changesets could not be found.

    In case the caller did a call possibly involving several changesets,
    the failing, perhaps more specific, expression can be set as argument.
    """



# Copyright 2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
import multiprocessing
import pytest
import signal
import socket
import sys
from socket import (
    AF_INET,
)
import time

from .. import prefork as server
from ..worker import WorkerProcess


def raiser(exc, *exc_args):
    def f(*a, **kw):
        raise exc(*(exc_args or ('something', )))

    return f


def test_catch_netloc_value_error(monkeypatch):
    # errors parsing the netloc
    monkeypatch.setattr(server, 'analyze_netloc', raiser(ValueError))
    url = "tcp://localhost"
    with pytest.raises(Exception) as exc_info:
        server.prebind_sockets([url]).__enter__()

    assert exc_info.value.args == (url, )


def test_so_portreuse_error(monkeypatch):
    monkeypatch.setattr(socket.socket, 'getsockopt', lambda *a: 0)
    url = "tcp://whatever"
    with pytest.raises(server.SocketReusePortError) as exc_info:
        server.prebind_sockets([url]).__enter__()
    assert exc_info.value.args == (url, )


def test_prebind_sockets(monkeypatch):
    bind_records = []

    def fake_bind(sock, port_info):
        bind_records.append(port_info)

    monkeypatch.setattr(socket.socket, 'bind', fake_bind)

    with server.prebind_sockets(['tcp://127.0.0.1:0',
                                 'tcp://[::1]:0',
                                 'tcp://localhost:0',
                                 'unix:/rel/path',
                                 ]):
        assert bind_records == [('127.0.0.1', 0),
                                ('::1', 0),
                                ('localhost', 0),
                                ]

    # a getsockopt that gives expected success values only for IPv4
    def getsockopt(sock, *a, **kw):
        if a[1] != socket.SO_REUSEPORT:  # pragma no cover (for compat)
            return socket.getsockopt(*a, **kw)
        return 0 if sock.family == AF_INET else 1

    monkeypatch.setattr(socket.socket, 'getsockopt', getsockopt)

    del bind_records[:]

    # will fail on the second, hence call the close() for the first
    # that doesn't fail and coverage tells us it's really called
    with pytest.raises(server.SocketReusePortError):
        server.prebind_sockets(['tcp://127.0.0.1:0',
                                'tcp://[::1]:0',
                                ]).__enter__()


def test_run_forever(monkeypatch, tmpdir):

    # server processes are subprocesses. We're only interested into
    # proper start/stop and correct arguments dispatching. Let's have
    # them write that to a file.
    workers_log = tmpdir.join('worker-start.log')

    def fake_server(wid, urls, storages, **kw):
        with open(workers_log, 'a') as logf:
            logf.write("%d %s\n" % (wid, ' '.join(urls)))

    def read_workers():
        res = {}
        for line in workers_log.readlines():
            split = line.split()
            wid = int(split[0])
            for url in split[1:]:
                res.setdefault(url, []).append(wid)
        workers_log.remove()
        return res

    monkeypatch.setattr(server, 'server_process', fake_server)

    tcp_url = 'tcp://localhost:1234'
    unix_url = 'unix:/hgitaly.socket'

    server.run_forever([tcp_url], {}, nb_workers=3, mono_process=True)
    workers = read_workers()
    assert sum(len(wids) for wids in workers.values()) == 1

    server.run_forever([tcp_url], {},
                       nb_workers=3,
                       monitoring_interval=0.01)
    workers = read_workers()
    assert len(workers[tcp_url]) == 3

    # we don't want more than one worker per unix URL because it's not
    # implemented yet
    server.run_forever([unix_url], {},
                       nb_workers=5,
                       monitoring_interval=0.01)

    workers = read_workers()
    assert len(workers[unix_url]) == 1

    # mixed scenario
    server.run_forever([unix_url, tcp_url], {},
                       nb_workers=5,
                       monitoring_interval=0.01)
    workers = read_workers()
    assert workers[unix_url] == [0]
    assert set(workers[tcp_url]) == {0, 1, 2, 3, 4}

    # defaulting based on CPU count
    server.run_forever([tcp_url], {}, None,
                       monitoring_interval=0.01)
    assert tcp_url in workers  # not so obvious if there's only one CPU

    # at least 2 unless we are on a single CPU system
    monkeypatch.setattr(multiprocessing, 'cpu_count', lambda: 2)
    assert len(workers[tcp_url]) >= 2


def test_terminate_workers(monkeypatch, tmpdir):

    def worker_callable(wid, queue, **kw):  # pragma no cover
        def bye(*a):
            queue.put((wid, 'bye'))
            sys.exit(0)

        signal.signal(signal.SIGTERM, bye)
        queue.put((wid, 'ready'))
        time.sleep(10)  # should be 100x times enough
        queue.put((wid, 'timeout'))  # pragma no cover

    queue = multiprocessing.Queue()

    def read_messages():
        msgs = {}
        for _ in range(2):
            msg = queue.get()
            msgs[msg[0]] = msg[1]
        return msgs

    workers = [WorkerProcess(process_args=(wid, queue),
                             process_callable=worker_callable)
               for wid in range(3)]

    for wp in workers:
        wp.init_process()

    # omitting to start a worker to trigger exception in its
    # termination, hence testing that we aren't impaired by that.
    for worker in workers[1:]:
        worker.start()

    # wait for all running workers to be ready to handle signal
    assert read_messages() == {wid: 'ready' for wid in range(1, 3)}

    shutdown_required = server.MutableBoolean(False)
    server.terminate_workers(shutdown_required, workers, signal.SIGTERM)
    for worker in workers[1:]:
        worker.join()

    assert shutdown_required  # flipped by terminate_workers()
    assert read_messages() == {wid: 'bye' for wid in range(1, 3)}


def test_mutable_boolean():
    mut_bool = server.MutableBoolean(False)
    assert bool(mut_bool) is False
    assert not mut_bool

    # demonstrate the main use-case: passing down to any callable
    # and getting back the changed value.
    def mutator(mutable, v):
        mutable.set(v)

    mutator(mut_bool, True)
    assert bool(mut_bool) is True
    assert mut_bool

    # getting back
    mutator(mut_bool, False)
    assert bool(mut_bool) is False
    assert not mut_bool



# Copyright 2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
import pytest
import time

import py.path
from hgext3rd.heptapod.branch import set_default_gitlab_branch
from hgext3rd.heptapod.special_ref import (
    special_refs,
)
from hgitaly.stub.shared_pb2 import (
    PaginationParameter,
)
from hgitaly.stub.ref_pb2 import (
    FindBranchRequest,
    FindDefaultBranchNameRequest,
    FindRefsByOIDRequest,
    FindTagError,
    FindTagRequest,
    FindLocalBranchesRequest,
    ListRefsRequest,
    DeleteRefsRequest,
    RefExistsRequest,
)
from hgitaly.stub.shared_pb2 import (
    SortDirection,
)
from hgitaly.stub.ref_pb2_grpc import RefServiceStub

from . import skip_comparison_tests
from .comparison import (
    normalize_commit_message,
)
if skip_comparison_tests():  # pragma no cover
    pytestmark = pytest.mark.skip

parametrize = pytest.mark.parametrize


def test_compare_find_branch(gitaly_comparison):
    fixture = gitaly_comparison
    git_repo = fixture.git_repo

    fixture.hg_repo_wrapper.write_commit('foo', message="Some foo")
    gl_branch = b'branch/default'

    # mirror worked
    assert git_repo.branch_titles() == {gl_branch: b"Some foo"}

    def normalize_response(rpc_helper, resp, **kw):
        normalize_commit_message(resp.branch.target_commit)

    rpc_helper = fixture.rpc_helper(
        stub_cls=RefServiceStub,
        method_name='FindBranch',
        request_cls=FindBranchRequest,
        response_sha_attrs=['branch.target_commit.id'],
        normalizer=normalize_response,
    )

    rpc_helper.assert_compare(name=gl_branch)

    # invalid case
    rpc_helper.assert_compare_errors(name=gl_branch, repository=None,
                                     same_details=False)


@parametrize('hg_server', ('hgitaly', 'rhgitaly'))
def test_compare_find_default_branch_name(gitaly_rhgitaly_comparison,
                                          hg_server):
    fixture = gitaly_rhgitaly_comparison
    git_repo = fixture.git_repo

    rpc_helper = fixture.rpc_helper(
        hg_server=hg_server,
        stub_cls=RefServiceStub,
        method_name='FindDefaultBranchName',
        request_cls=FindDefaultBranchNameRequest,
    )

    # empty repo
    rpc_helper.assert_compare()

    fixture.hg_repo_wrapper.write_commit('foo', message="Some foo")
    gl_branch = b'branch/default'

    # mirror worked
    assert git_repo.branch_titles() == {gl_branch: b"Some foo"}

    rpc_helper.assert_compare()
    # invalid case
    rpc_helper.assert_compare_errors(repository=None,
                                     same_details=False)


@parametrize('hg_server', ('hgitaly', 'rhgitaly'))
def test_compare_find_default_branch_name_fallbacks(
        gitaly_rhgitaly_comparison,
        hg_server
):
    fixture = gitaly_rhgitaly_comparison
    git_repo = fixture.git_repo

    rpc_helper = fixture.rpc_helper(
        hg_server=hg_server,
        stub_cls=RefServiceStub,
        method_name='FindDefaultBranchName',
        request_cls=FindDefaultBranchNameRequest,
    )

    # empty repo, but with default branch set
    branch_name = b'branch/maybe_later'
    set_default_gitlab_branch(fixture.hg_repo_wrapper.repo, branch_name)
    # git_repo.set_symref needs a `py.path` path, not a `pathlib.Path`
    git_repo = fixture.git_repo
    git_repo.path = py.path.local(git_repo.path)
    git_repo.set_symref('HEAD', 'refs/heads/' + branch_name.decode('ascii'))
    rpc_helper.assert_compare()

    # non-empty repo, Gitaly will return the first existing branch it finds
    # (actually works because mirroring/state_maintainer fixes the default
    # branch to be an existing one)
    fixture.hg_repo_wrapper.write_commit('foo', message="Some foo")
    gl_branch = b'branch/default'

    # mirror worked
    assert git_repo.branch_titles() == {gl_branch: b"Some foo"}
    rpc_helper.assert_compare()


@parametrize('hg_server', ('hgitaly', 'rhgitaly'))
def test_compare_ref_exists(gitaly_rhgitaly_comparison, hg_server):
    fixture = gitaly_rhgitaly_comparison
    wrapper = fixture.hg_repo_wrapper

    rpc_helper = fixture.rpc_helper(
        hg_server=hg_server,
        stub_cls=RefServiceStub,
        method_name='RefExists',
        request_cls=RefExistsRequest,
        )

    assert_compare = rpc_helper.assert_compare
    assert_compare_errors = rpc_helper.assert_compare_errors

    # empty repo, when state files don't exist yet
    assert_compare(ref=b'refs/heads/branch/default')
    assert_compare(ref=b'refs/keep-around/' + b'1234' * 10)

    ctx0 = wrapper.commit_file('foo')
    wrapper.command('tag', b'start-tag', rev=ctx0.hex())
    ctx1 = wrapper.commit_file('foo', topic='sampletop')
    fixture.invalidate()

    mr_ref_path, _ = fixture.write_special_ref(b'merge-requests/2/train',
                                               ctx1.hex())
    hg_ka_ref_path, git_ka_ref_path = fixture.create_keep_around(ctx1.hex())

    assert_compare(ref=b'refs/heads/branch/default')
    assert_compare(ref=b'refs/heads/topic/default/sampletop')
    assert_compare(ref=b'refs/heads/not-found')
    assert_compare(ref=b'refs/tags/start-tag')
    assert_compare(ref=b'refs/tags/unknown-tag')
    assert_compare(ref=b'refs/merge-requests/2/train')
    assert_compare(ref=b'refs/merge-requests/3/head')  # result is False
    assert_compare(ref=b'refs/')
    assert_compare(ref=b'refs/tags')
    assert_compare(ref=b'refs/tags/')
    # assert_compare() not being able to convert the keep-around ref path from
    # Mercurial to Git on the fly, we need to go lower level
    hg_resp = rpc_helper.rpc('hg', ref=hg_ka_ref_path)
    git_resp = rpc_helper.rpc('git', ref=git_ka_ref_path)
    assert hg_resp == git_resp

    assert_compare(ref=b'refs/keep-around/not-even-a-hash')

    assert_compare_errors(ref=b'refs')
    assert_compare_errors(ref=b'HEAD')
    assert_compare_errors(ref=b'notrefs/something')


def test_compare_find_local_branches(gitaly_comparison):
    fixture = gitaly_comparison
    wrapper = fixture.hg_repo_wrapper

    # make three branches with the 3 possible orderings differ
    now = time.time()
    commit_ages = {0: 30, 1: 40, 2: 20}
    for i in range(3):
        wrapper.commit_file('foo', branch='br%02d' % i, return_ctx=False,
                            utc_timestamp=now - commit_ages[i])
    # mirror worked
    assert set(fixture.git_repo.branch_titles().keys()) == {
        b'branch/br%02d' % i for i in range(3)}

    def normalize_response(rpc_helper, resp, **kw):
        for chunk in resp:
            for branch in chunk.local_branches:
                normalize_commit_message(branch.target_commit)

    rpc_helper = fixture.rpc_helper(
        stub_cls=RefServiceStub,
        method_name='FindLocalBranches',
        request_cls=FindLocalBranchesRequest,
        streaming=True,
        response_sha_attrs=['local_branches[].target_commit.id',
                            'local_branches[].target_commit.parent_ids[]',
                            ],
        normalizer=normalize_response,
    )

    def assert_compare(limit=0, page_token='', pagination=True, **kw):
        if pagination:
            pagination_params = PaginationParameter(limit=limit,
                                                    page_token=page_token)
        else:
            pagination_params = None

        rpc_helper.assert_compare(pagination_params=pagination_params, **kw)

    for limit in (0, 3, 8, -1):
        assert_compare(limit=limit)

    # case without any pagination parameters
    assert_compare(123, pagination=False)

    assert_compare(10, page_token='refs/heads/branch/br01')

    # sort options
    for sort_by in FindLocalBranchesRequest.SortBy.values():
        assert_compare(10, sort_by=sort_by)


@parametrize('hg_server', ('hgitaly', 'rhgitaly'))
def test_find_tag(gitaly_rhgitaly_comparison, hg_server):
    fixture = gitaly_rhgitaly_comparison
    hg_wrapper = fixture.hg_repo_wrapper

    hg_wrapper.commit_file('foo')
    hg_wrapper.command('tag', b'v3.3', rev=b'.')
    hg_wrapper.command('gitlab-mirror')

    # mirror worked
    assert fixture.git_repo.tags() == {b'v3.3'}

    def normalize_response(rpc_helper, resp, **kw):
        normalize_commit_message(resp.tag.target_commit)
        # no tag ids in Mercurial until we can assign a tagging changeset
        # (requires a long-term core change)
        resp.tag.id = ''

    rpc_helper = fixture.rpc_helper(
        hg_server=hg_server,
        stub_cls=RefServiceStub,
        method_name='FindTag',
        response_sha_attrs=['tag.target_commit.id'],
        request_cls=FindTagRequest,
        normalizer=normalize_response,
    )
    rpc_helper.assert_compare(tag_name=b'v3.3')

    # Fine structured errors comparison even though they are expected to
    # be identical, for coverage of comparison method in `RpcHelper`:
    errors_handler = dict(git_cls=FindTagError, to_git=lambda e: e)
    rpc_helper.assert_compare_errors(tag_name=b'nosuchtag',
                                     structured_errors_handler=errors_handler)


def test_delete_refs(gitaly_comparison):
    fixture = gitaly_comparison
    git_repo = fixture.git_repo
    hg_wrapper = fixture.hg_repo_wrapper

    base_hg_ctx = hg_wrapper.commit_file('foo')
    hg_sha = base_hg_ctx.hex()
    mr_ref_name = b'merge-requests/2/train'
    mr_ref_path, git_sha = fixture.write_special_ref(mr_ref_name, hg_sha)

    rpc_helper = fixture.rpc_helper(stub_cls=RefServiceStub,
                                    method_name='DeleteRefs',
                                    request_cls=DeleteRefsRequest)

    assert_compare = rpc_helper.assert_compare
    assert_compare_errors = rpc_helper.assert_compare_errors

    assert_compare_errors(refs=[b'xy'], except_with_prefix=[b'refs/heads'])
    assert_compare(refs=[mr_ref_path])

    # unknown refs dont create errors
    fixture.write_special_ref(mr_ref_name, hg_sha)
    unknown = b'refs/environments/imaginary'
    assert_compare(refs=[unknown])

    # also mixing unknown with known is ok
    assert_compare(refs=[unknown, mr_ref_path])

    assert git_repo.all_refs() == {b'refs/heads/branch/default': git_sha}
    hg_wrapper.reload()
    assert special_refs(hg_wrapper.repo) == {}

    # using except_with_prefix
    env_ref_name = b'environments/2877'
    env_ref_path, _ = fixture.write_special_ref(env_ref_name, hg_sha)

    # on the Mercurial side, we'll consider the special ref only,
    # but on the Git side, the `refs/heads` prefix has to be ignored.
    # This is similar to what the current actual caller,
    # `Projects::AfterImportService`, does.
    for except_prefixes in (
            [b'refs/environments/', b'refs/heads/'],
            [b'refs/environments', b'refs/heads/'],
            [b'refs/envir', b'refs/heads/'],
            ):
        fixture.write_special_ref(mr_ref_name, hg_sha)
        fixture.write_special_ref(env_ref_name, hg_sha)

        assert_compare(except_with_prefix=except_prefixes)
        assert git_repo.all_refs() == {b'refs/heads/branch/default': git_sha,
                                       env_ref_path: git_sha}
        hg_wrapper.reload()
        assert special_refs(hg_wrapper.repo) == {env_ref_name: hg_sha}


@parametrize('hg_server', ('hgitaly', 'rhgitaly'))
def test_list_refs(gitaly_rhgitaly_comparison, hg_server):
    fixture = gitaly_rhgitaly_comparison
    hg_wrapper = fixture.hg_repo_wrapper

    def normalize_refs(rpc_helper, resps, vcs=None):
        if vcs != 'hg':
            return

        prefix = b'refs/keep-around/'

        for resp in resps:
            all_pseudo_ref_idx = None
            for i, ref_msg in enumerate(resp.references):
                if ref_msg.name == b'ALL' and rpc_helper.norm_all_pseudo_ref:
                    all_pseudo_ref_idx = i
                if ref_msg.name.startswith(prefix):
                    ref_msg.name = prefix + rpc_helper.hg2git(
                        ref_msg.name[len(prefix):])
            if all_pseudo_ref_idx is not None:
                del resp.references[all_pseudo_ref_idx]

    rpc_helper = fixture.rpc_helper(
        hg_server=hg_server,
        stub_cls=RefServiceStub,
        method_name='ListRefs',
        request_cls=ListRefsRequest,
        request_defaults=dict(patterns=[b"refs/"], head=True),
        streaming=True,
        request_sha_attrs=['pointing_at_oids'],
        response_sha_attrs=['references[].target'],
        normalizer=normalize_refs,
    )
    rpc_helper.norm_all_pseudo_ref = True
    # empty repo, in particular no default GitLab branch (HEAD)
    rpc_helper.assert_compare()

    # make three changesets on which the 3 possible orderings differ
    # (in Mercurial committer and author dates are the same)
    now = time.time()
    commit_ages = {0: 30, 1: 40, 2: 20}
    hg_sha = hg_wrapper.commit_file('bar').hex()

    # branches
    hg_shas = []
    for i in range(3):
        hg_shas.append(
            hg_wrapper.commit_file('foo', branch='br%02d' % i,
                                   utc_timestamp=now - commit_ages[i]).hex()
        )
    rpc_helper.assert_compare()
    for head in False, True:
        rpc_helper.assert_compare(patterns=[b'refs/heads/'], head=head)

    # with tags
    hg_wrapper.command('tag', b'v3.1', rev=b'br01')
    hg_wrapper.command('gitlab-mirror')
    fixture.invalidate()
    rpc_helper.assert_compare()
    for head in False, True:
        rpc_helper.assert_compare(patterns=[b'refs/heads/'], head=head)
        rpc_helper.assert_compare(patterns=[b'refs/tags/'], head=head)

    # now with special refs
    fixture.write_special_ref(b'pipeline/13', hg_sha)
    rpc_helper.assert_compare()
    for head in False, True:
        rpc_helper.assert_compare(patterns=[b'refs/heads/'], head=head)
        rpc_helper.assert_compare(patterns=[b'refs/tags/'], head=head)

    # with a keep around
    fixture.create_keep_around(hg_sha)
    rpc_helper.assert_compare()
    for head in False, True:
        rpc_helper.assert_compare(patterns=[b'refs/heads/'], head=head)
        rpc_helper.assert_compare(patterns=[b'refs/tags/'], head=head)

    # more general patterns
    rpc_helper.norm_all_pseudo_ref = False
    for head in False, True:
        # notice how HEAD is ignored by pattern matching
        rpc_helper.assert_compare(patterns=[b'refs/heads/branch/br*'],
                                  head=head)

        # 'branch/something' should not match `br*` and special case of
        # HEAD not being part of response if nothing matches
        rpc_helper.assert_compare(patterns=[b'refs/heads/br*'], head=head)

    # trailing slash is stricly prefixing (no wildcards)
    rpc_helper.assert_compare(patterns=[b'refs/h*s/'])

    # if wildcards are interpreted, they can be in any segment, yet do not
    # match slashes and need equal numbers of segment (no prefixing by pattern)
    rpc_helper.assert_compare(patterns=[b'refs/h*s/branch/br*'])  # matching
    rpc_helper.assert_compare(patterns=[b'refs/*/br*'])  # not matching slash
    rpc_helper.assert_compare(patterns=[b'refs/*/branch'])  # prefixing case

    # logical conjunction of several patterns
    rpc_helper.assert_compare(patterns=[b'refs/heads/branch/br*',
                                        b'refs/heads/branch/d*'],
                              head=head)

    # no error that we can think of would be specific of this gRPC method

    # pointing_at_oids
    for head in False, True:
        rpc_helper.assert_compare(pointing_at_oids=[hg_shas[0]], head=head)
        rpc_helper.assert_compare(pointing_at_oids=[hg_shas[0], hg_shas[1]],
                                  head=head)
        # together with patterns
        for i in range(3):
            # empty result if `i == 0 and not head`
            rpc_helper.assert_compare(pointing_at_oids=[hg_shas[i]],
                                      patterns=[b'refs/tags/'],
                                      head=head)

    # sorting by date (not so important to be exactly as in Gitaly, we have
    # less date fields anyway, but the point is to test that RHGitaly works)
    rpc_helper.norm_all_pseudo_ref = True
    SortBy = ListRefsRequest.SortBy
    for direction in (SortDirection.ASCENDING, SortDirection.DESCENDING):
        rpc_helper.assert_compare(
            sort_by=SortBy(key=SortBy.AUTHORDATE,
                           direction=direction),
            patterns=[b"refs/heads/branch/br*"],
            head=False,
        )

    # FindRefsByOID is almost a subset of ListRefs, the only stated
    # thing that ListRefs would not do is accepting oids by prefix.
    # So we'll use the same setup

    rpc_helper = fixture.rpc_helper(
        stub_cls=RefServiceStub,
        method_name='FindRefsByOID',
        request_cls=FindRefsByOIDRequest,
        request_defaults=dict(ref_patterns=["refs/"]),
        request_sha_attrs=['oid'],
    )

    hg_short_shas = [sha[:12] for sha in hg_shas]
    git_short_shas = [rpc_helper.hg2git(sha)[:12] for sha in hg_shas]
    fixture.hg_git._map_hg.update(zip(hg_short_shas, git_short_shas))

    # no need to proceed further if this fails:
    assert rpc_helper.hg2git(hg_short_shas[1]) == git_short_shas[1]

    for hg_sha in hg_shas + hg_short_shas:
        for patterns in (
                ['refs/tags/'],
                ['refs/heads/'],
                [],
                ):
            rpc_helper.assert_compare(oid=hg_sha.decode(),
                                      ref_patterns=patterns)

    # hg_shas[1] has two refs (a branch and a tag)
    for limit in (0, 1, 2):
        rpc_helper.assert_compare(oid=hg_shas[1].decode(), limit=limit)



# Copyright 2022-2023 Georges Racinet <georges.racinet@octobus.net>
# Copyright 2024 Georges Racinet <georges.racinet@cloudcrane.io>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
import argparse
from collections import defaultdict
from datetime import datetime
import json
import re
import statistics
import sys

ISOFORMAT_WITH_TZ = sys.version_info >= (3, 11)

RHGITALY_SPAN_CLOSE_RX = re.compile(
    r'(\d+-\d+-\d+T\d+:\d+:\d+\.\d+Z).*?(?:\[1m)?(\w+)(.\[0m.\[1m)?'
    r'\{.*rhgitaly::service.*? '
    r'close .*?time.busy.*?=.*?([0-9.]*?)([mnµ]?)s ')

REQUEST_SUFFIX = 'Request'

unit_multipliers_to_millis = {
    '': 1000,
    'm': 1,
    'µ': 1e-3,
    'n': 1e-6,
}


def hgitaly_parse_logs(fobj, since=None):  # pragma no cover
    """Parse a log file in HGitaly format.

    :returns: (first timestamp, call counts per request, execution times)
       Call counts and execution times are per requests, execution times
       only for requests with normal termination (i.e, not cancelled or
       failed)
    """
    called = defaultdict(lambda: 0)
    succeeded = defaultdict(list)
    first_timestamp = None
    for line in fobj:
        try:
            log = json.loads(line)
        except json.JSONDecodeError:
            # ignoring log lines predating the transition to JSON
            continue

        if first_timestamp is None:
            log_dt = log['asctime']
            if since is not None:
                if ISOFORMAT_WITH_TZ and not log_dt.endswith('Z'):
                    # the whole HGitaly process timezone is UTC, but this
                    # may not be explicit in the logging format
                    log_dt += 'Z'
                if not ISOFORMAT_WITH_TZ and log_dt.endswith('Z'):
                    # this Python version does not support time zones in
                    # ISO format, let's go the naive way.
                    log_dt = log_dt[:-1]
                # some Python versions accept only dots for subsecond part
                dt = datetime.fromisoformat(log_dt.replace(',', '.'))
                if dt < since:
                    continue
            first_timestamp = log['asctime']

        req = log.get('request')
        if req is None:
            continue

        req = req.split(' ', 1)[0]
        if req.endswith(REQUEST_SUFFIX):
            req = req[:-len(REQUEST_SUFFIX)]
        msg = log.get('message')
        if msg.startswith("Starting"):
            called[req] += 1
        elif msg.startswith("Finished"):
            # milliseconds for readability of most frequent cases
            succeeded[req].append(log["elapsed_seconds"] * 1000)

    return first_timestamp, called, succeeded


def snake_case_to_camel(s):
    """Change case of key from snake_case to CamelCase.

    >>> snake_case_to_camel('find_commits')
    'FindCommits'
    """
    return ''.join(w.capitalize() for w in s.split('_'))


def dict_snake_case_to_camel(d):
    """Change in place the case of keys.

    >>> d = dict(hg_publish=3)
    >>> dict_snake_case_to_camel(d)
    >>> d
    {'HgPublish': 3}
    """
    keys = tuple(d.keys())
    for k in keys:
        v = d.pop(k)
        d[snake_case_to_camel(k)] = v


def rhgitaly_parse_logs(fobj, since=None):  # pragma no cover
    """Parse a log file in HGitaly format.

    Same as :func:`hgitaly_parse_logs`, for RHGitaly logs format, except that
    we return `None` as `first_timestamp` (analysis not done yet, TODO)
    """
    called = defaultdict(lambda: 0)
    succeeded = defaultdict(list)
    first_timestamp = None
    for line in fobj:
        m = RHGITALY_SPAN_CLOSE_RX.search(line)

        if m is None:
            continue
        if first_timestamp is None:
            log_dt = m.group(1)
            if since is not None:
                # RHGitaly log date/times are always in full ISO format
                # (with 'T' and 'Z'). Some Python versions don't understand
                # the use of commas, though
                dt = datetime.fromisoformat(log_dt.replace(',', '.'))
                if dt < since:
                    continue
            first_timestamp = log_dt
        req = m.group(2)
        called[req] += 1
        # TODO dectect request failures
        succeeded[req].append(
            float(m.group(4)) * unit_multipliers_to_millis[m.group(5)])
        continue

    dict_snake_case_to_camel(called)
    dict_snake_case_to_camel(succeeded)
    return first_timestamp, called, succeeded


def stats():  # pragma no cover
    parser = argparse.ArgumentParser(
        description="Analyse HGitaly logs and output statistics by method",
    )
    parser.add_argument('log_path',
                        help="Path to the log file to analyse. "
                        "Use `-` for `stdin`")
    parser.add_argument('--json', action='store_true')
    parser.add_argument('--rhgitaly', action='store_true')
    parser.add_argument('--since',
                        help="Filter the logs from the given date, "
                        "that must be in UTC ISO-8601 format with time zone")
    parser.add_argument('--sort-by', choices=('calls',
                                              'incomplete',
                                              'timing:mean',
                                              'timing:median',
                                              'timing:worst_decile',
                                              'timing:worst_centile',
                                              ),
                        default='calls')

    cl_args = parser.parse_args()
    log_path = cl_args.log_path
    rhgitaly = cl_args.rhgitaly

    called = defaultdict(lambda: 0)
    succeeded = defaultdict(list)

    if rhgitaly:
        parse = rhgitaly_parse_logs
    else:
        parse = hgitaly_parse_logs

    if log_path == '-':
        logf = sys.stdin
    else:
        logf = open(log_path)

    if cl_args.since is None:
        since = None
    else:
        try:
            since = datetime.fromisoformat(cl_args.since)
        except ValueError:
            parser.error("--since must be given in ISO-8601 format")
        if since.tzinfo is None and ISOFORMAT_WITH_TZ:
            parser.error("--since must be given with timezone, "
                         "usually just 'Z' for UTC")
    with logf:
        first_timestamp, called, succeeded = parse(logf, since=since)

    total_requests = sum(called.values())
    incomplete = total_requests - sum(len(succ) for succ in succeeded.values())

    stats = []
    for req, count in called.items():
        percent = round(count * 100 / total_requests)
        succ = succeeded[req]
        nb_succ = len(succ)
        req_stats = dict(percent=percent,
                         calls=count,
                         complete=nb_succ,
                         incomplete=count - nb_succ,
                         )
        stats.append((req, req_stats))
        if nb_succ > 0:
            timing = req_stats['completion_stats_ms'] = dict(
                mean=statistics.mean(succ),
                median=statistics.median(succ),
            )
            if nb_succ > 1:
                timing['standard_deviation'] = statistics.pstdev(succ)
                deciles = statistics.quantiles(succ, method='inclusive', n=10)
                centiles = statistics.quantiles(succ, method='inclusive',
                                                n=100)
                timing['best_centile'] = centiles[0]
                timing['best_decile'] = deciles[0]
                timing['worst_decile'] = deciles[-1]
                timing['worst_centile'] = centiles[-1]

    sort_key = cl_args.sort_by
    if sort_key.startswith('timing:'):
        sort_key = ('completion_stats_ms', sort_key.split(':', 1)[-1])

    def sort_key_function(req_stats):
        stats = req_stats[1]
        if isinstance(sort_key, str):
            return -stats.get(sort_key, 0)

        stats = stats.get(sort_key[0])
        if stats is None:
            return 0
        return -stats.get(sort_key[1], 0)

    stats.sort(key=sort_key_function)
    stats = dict(stats)

    if cl_args.json:
        print(json.dumps(stats))
        return

    print(f"TOTAL Requests since {first_timestamp}: {total_requests} \n"
          f"      {incomplete} incomplete (cancelled or failed)")
    print("Breakdown:")
    for req, details in stats.items():
        percent = '%4.1f' % details['percent']
        time_stats = details.get('completion_stats_ms')
        if time_stats is None:
            print(f"  {percent}% {req} "
                  f"(called={details['calls']}, "
                  f"incomplete={details['incomplete']} (timing N/A)")
            continue
        time_stats = details['completion_stats_ms']
        mean = time_stats['mean']
        avg_ms_str = '%.1f' % mean
        worst_decile_str = '%.1f' % time_stats.get('worst_decile', mean)
        print(f"  {percent}% {req} "
              f"(called={details['calls']}, "
              f"incomplete={details['incomplete']}, "
              f"average_time={avg_ms_str}ms, "
              f"worst_decile={worst_decile_str}ms)")



# Copyright 2023 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
from contextlib import contextmanager
import attr
import errno
import logging
import os
from pathlib import Path
import shutil
import time
from typing import Any
import weakref

from mercurial import (
    cmdutil,
    commands,
    error,
    hg,
    lock,
)
from mercurial.repoview import _filteredrepotypes

from .procutil import is_current_service_process
from .revision import gitlab_revision_changeset

logger = logging.getLogger(__name__)

ROSTER_FILE_NAME = b'hgitaly-working-dirs'
ROSTER_LOCK_NAME = ROSTER_FILE_NAME + b'.lock'


def clear_repo_class(repo_class):
    _filteredrepotypes.pop(repo_class, None)


def wd_path(per_repo_root, wd_id):
    return Path(f'{per_repo_root}/{wd_id}')


class WorkingDirectoryError(RuntimeError):
    """Represent an error obtaining a working directory.

    The first argument is the type of action, currently one of
    ``(load, create, update)``.
    The second argument is the original exception
    """


@attr.define
class WorkingDirectory:
    """Represents a semi-permanent working directory."""
    id: int
    branch: bytes
    path: Path
    repo: Any = attr.field(repr=False, default=None)

    def roster_line(self, pid=None, timestamp=None):
        """Return a line suitable to put in the roster file.

        :param pid: use ``None`` to give back the working directory to the
          pool, current process id to reserve it.
        :return: bytes, since this is to write in files managed by Mercurial's
         vfs, with end of line included, suitable for use in ``writelines``
         and ``write`` methods.
        """
        if timestamp is None:
            timestamp = int(time.time())

        pid_bytes = b'-' if pid is None else str(pid).encode('ascii')
        return (b'%d %s %d %s\n' % (self.id, pid_bytes,
                                    timestamp, self.branch))

    def load_or_create(self, ui, src_repo):
        path_bytes = bytes(self.path)
        if self.path.exists():
            try:
                self.init_repo(hg.repository(ui, path_bytes))
            except error.Abort as exc:
                raise WorkingDirectoryError('load', exc)
        else:
            try:
                self.init_repo(hg.share(ui, src_repo.root, path_bytes,
                                        update=False))
            except error.Abort as exc:
                raise WorkingDirectoryError('create', exc)

    def command(self, name, *args, hidden=False, **kwargs):
        cmd = cmdutil.findcmd(name, commands.table)[1][0]
        repo = self.repo
        if hidden:
            repo = repo.unfiltered()
        return cmd(repo.ui, repo, *args, **kwargs)

    def clean_update(self, target_rev):
        try:
            self.uncaught_update(target_rev, clean=True, hidden=True)
        except error.Abort as exc:
            raise WorkingDirectoryError('update', exc)

    def uncaught_update(self, target_rev, **kw):
        self.command(b'update', target_rev, **kw)

    def purge(self):
        self.command(b'purge', all=True, hidden=True, confirm=False)

    def init_repo(self, repo):
        weakref.finalize(repo, clear_repo_class, repo.unfiltered().__class__)
        self.repo = repo
        # any existing wlock is lingering, since the whole point of
        # this module is to provide a higher level locking system
        repo.vfs.tryunlink(b'wlock')


@contextmanager
def working_directory(workdirs_root, repo, changeset=None):
    """Context manager for temporary working directories

    Entering this context manager, (typically with the ``with`` statement
    yields a clean working directory for the given repository, cleanly updated
    at the specified changeset.

    The working directory is represented by a :class:`WorkingDirectory`
    instance and is guaranteed to be for exclusive use by the caller.
    It is relturned to the underlying pool when exiting the context.

    The changeset Mercurial branch is used as a heuristic for reuse: it
    is assumed that a working directory previously updated to a changeset
    with the same branch can be quickly updated to the given one.

    The given changeset can be obsolete, allowing to revive it with more than
    a simple `hg touch` equivalent. It is therefore up to the caller to
    check for obsolescence if undesireable (e.g., in mergeability checks).

    :param workdirs_root: path to the directory where working directories
      for the given repo are kept.
    :param repo: Mercurial repository object
    :param changeset: a :class:`changectx` instance. Use `None` to get an
      empty working directory. It will in this case be considered to be for
      the ``default`` branch. This should be used only in the case of
      empty repositories.
    """
    branch = b'default' if changeset is None else changeset.branch()
    wd = reserve_workdir(workdirs_root, repo.ui, repo, branch)
    wd.load_or_create(repo.ui, repo)
    wd.purge()
    if changeset is not None:
        wd.clean_update(changeset.hex())

    try:
        yield wd
    finally:
        release_workdir(repo, wd)


class rosterlock(lock.lock):

    def _lockshouldbebroken(self, locker):
        pid = int(locker.split(b":", 1)[-1].decode('ascii').strip())
        return not is_current_service_process(pid)


def trylock(ui, vfs, lockname, timeout, warntimeout, *args, step=0.1,
            **kwargs):
    """return an acquired lock or raise an a LockHeld exception

    This function is responsible to issue logs about
    the held lock while trying to acquires it.

    Derived from Mercurial's `lock.trylock`, with these differences:

    - using :data:`logger`
    - using :class:`rosterlock`
    - configurable sleep interval (param ``step``), both because roster
      file operations are designed to be fast, and to reach the warntimeout
      in tests without overlong sleeps.
    - timeouts always exist, since server operation must not be stalled
      forever.
    - `acquirefn` is ignored, as we don't need it for the roster lock.
    """

    def log(level, locker):
        """log the "waiting on lock" message through at the given level"""
        pid = locker.split(b":", 1)[-1].decode('ascii')
        logger.log(level, "waiting for lock on %r held by process %r",
                   lk.desc, pid)

    lk = rosterlock(vfs, lockname, 0, *args, dolock=False, **kwargs)

    debugidx = 0
    warningidx = int(warntimeout / step)

    delay = 0
    while True:
        try:
            lk._trylock()
            break
        except error.LockHeld as inst:
            if delay == debugidx:
                log(logging.DEBUG, inst.locker)
            if delay == warningidx:
                log(logging.WARNING, inst.locker)
            if timeout <= delay * step:
                raise error.LockHeld(
                    errno.ETIMEDOUT, inst.filename, lk.desc, inst.locker
                )
            time.sleep(step)
            delay += 1

    lk.delay = delay
    if delay:
        if 0 <= warningidx <= delay:
            logger.warning("got lock after %.2f seconds", delay * step)
        else:
            logger.debug("got lock after %.2f seconds", delay * step)

    return lk


@contextmanager
def locked_roster(repo, timeout=1):
    """Locked read/write access to the repo working directories roster file.

    This context manager provides a pair of open files.
    The first is to be used to read the roster, the second to write it,
    with atomic update occurring at the end of the context.

    The lock is not reentrant, which is good enough for this simple need of
    a very short-lived lock protecting both readind and writing.

    :param timeout: maximum time to wait until the roster lock is acquired.
      The default value is intended for production, tests will set it to
      shorter values.
    """
    vfs = repo.vfs
    # TODO Mercurial lock system does not allow to customize its breaking
    # logic, which is actually causing deadlocks in containers.
    # In HGitaly, we can have much more certainty because of the general
    # prerequisites that a HGitaly service (typically several processes)
    # has exclusive access to this resource.
    # This is more general than the working directories roster lock, but
    # it is dubious that HGitaly ever gets exclusive access to Mercurial
    # content (HTTP pushes could be handled if HGitaly eventually manages
    # the hgwebdir services, but SSH pushes would not).
    ui = repo.ui
    warntimeout = 3 * timeout / 10
    # internal config: ui.signal-safe-lock
    signalsafe = ui.configbool(b'ui', b'signal-safe-lock')

    with trylock(ui, vfs,
                 lockname=ROSTER_LOCK_NAME,
                 timeout=timeout,
                 warntimeout=warntimeout,
                 releasefn=None,
                 acquirefn=None,
                 desc=b'Working directories roster file for %s' % repo.root,
                 signalsafe=signalsafe,
                 step=timeout / 10,
                 ):
        try:
            inf = vfs(ROSTER_FILE_NAME, b'rb')
        except FileNotFoundError:
            inf = None

        with vfs(ROSTER_FILE_NAME, b'wb', atomictemp=True) as outf:
            try:
                yield inf, outf
            finally:
                if inf is not None:
                    inf.close()


def roster_iter(fobj):
    if fobj is None:
        return

    for line in fobj:
        # id, pid, branch  TODO what to do in case of invalid data?
        wd_id, pid, timestamp, wd_branch = line.split(b' ', 3)
        pid = int(pid) if pid != b'-' else None
        yield (int(wd_id), pid, int(timestamp), wd_branch.strip()), line


def roster_branch_match(fobj, branch):
    matching = None
    max_id = - 1
    lines = []
    for (wd_id, pid, _ts, wd_branch), line in roster_iter(fobj):
        if wd_id > max_id:
            max_id = wd_id
        if wd_branch == branch:
            if pid is None or not is_current_service_process(pid):
                matching = wd_id
                continue
        lines.append(line)
    return matching, max_id + 1, lines


def reserve_workdir(workdirs_root, ui, repo, branch):
    current_pid = os.getpid()
    with locked_roster(repo) as (rosterf, new_rosterf):
        matching_id, unused_id, other_lines = roster_branch_match(rosterf,
                                                                  branch)
        wd_id = unused_id if matching_id is None else matching_id
        wd = WorkingDirectory(id=wd_id,
                              branch=branch,
                              path=wd_path(workdirs_root, wd_id),
                              )
        new_rosterf.writelines(other_lines)
        new_rosterf.write(wd.roster_line(pid=current_pid))
    return wd


def release_workdir(repo, wd):
    with locked_roster(repo) as (rosterf, new_rosterf):
        for parsed, line in roster_iter(rosterf):
            if parsed[0] != wd.id:
                new_rosterf.write(line)
            new_rosterf.write(wd.roster_line(pid=None))


def workdirs_gc(workdirs_root, repo, max_age_seconds, now=None):
    """Purge working directories that have not been used for long.

    At least one working directory for the GitLab default branch is
    always kept.
    If later on we start seeding working directories from other working
    directories, even if branches differ (it could indeed be faster on
    average), then we might want to always keep at least one, just preferably
    for the default branch if possible.

    In case a removal fails, the corresponding roster line is kept for
    consistency, and so that subsequent allocation does not end in error
    (assuming the disk is working again anyway).

    :param max_age: time, in seconds since last time the working directory
       was used.
    :param now: current time in seconds since Unix epoch (should be used in
       tests only)
    """
    if now is None:
        now = time.time()

    # using gitlab_revision_changeset spares us the back-and-forth between
    # GitLab and Mercurial branches, with its special cases (topic…)
    head = gitlab_revision_changeset(repo, b'HEAD')
    default_branch = None if head is None else head.branch()
    default_branch_kept = False
    to_remove = {}
    to_keep = []
    with locked_roster(repo) as (rosterf, new_rosterf):
        for (wd_id, pid, timestamp, branch), line in roster_iter(rosterf):
            if pid is None and now - timestamp > max_age_seconds:
                to_remove.setdefault(branch, []).append((wd_id, line))
            else:
                to_keep.append(line)
                if branch == default_branch:
                    default_branch_kept = True

        if not default_branch_kept:
            default_branch_to_rm = to_remove.get(default_branch)
            if default_branch_to_rm:
                to_keep.append(default_branch_to_rm.pop()[1])

        for _, to_rm in to_remove.items():
            for wd_id, line in to_rm:
                path = wd_path(workdirs_root, wd_id)
                try:
                    shutil.rmtree(path)
                except Exception:
                    logger.exception("Failed to remove working directory '%s' "
                                     "for repo at %r", path, repo.root)
                    to_keep.append(line)

        new_rosterf.writelines(to_keep)


def remove_all_workdirs(workdirs_root, repo):
    """Remove all workdirs for the given repository.

    This is to be used when preparing to remove the repository itself.
    """
    with locked_roster(repo) as (rosterf, new_rosterf):
        remove_all_workdirs_bare(workdirs_root)
        # new_rosterf being empty, this will void the existing one.


def remove_all_workdirs_bare(workdirs_root):
    """Removal of workdirs_root with no roster management.

    To be used either with roster being managed (lock acquired by caller etc.)
    or if the roster does not matter any more (repository is being entirely
    removed).

    Does not raise exceptions, as the caller typically will prefer to proceed
    with other cleanups if possible.
    """
    try:
        if os.path.exists(workdirs_root):
            shutil.rmtree(workdirs_root)
    except Exception:
        logger.exception("Failed to remove working directories at %r",
                         workdirs_root)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: blob.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import lint_pb2 as lint__pb2
from . import shared_pb2 as shared__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\nblob.proto\x12\x06gitaly\x1a\nlint.proto\x1a\x0cshared.proto\"Z\n\x0eGetBlobRequest\x12,\n\nrepository\x18\x01 \x01(\x0b\x32\x12.gitaly.RepositoryB\x04\x98\xc6,\x01\x12\x0b\n\x03oid\x18\x02 \x01(\t\x12\r\n\x05limit\x18\x03 \x01(\x03\":\n\x0fGetBlobResponse\x12\x0c\n\x04size\x18\x01 \x01(\x03\x12\x0c\n\x04\x64\x61ta\x18\x02 \x01(\x0c\x12\x0b\n\x03oid\x18\x03 \x01(\t\"\xbc\x01\n\x0fGetBlobsRequest\x12,\n\nrepository\x18\x01 \x01(\x0b\x32\x12.gitaly.RepositoryB\x04\x98\xc6,\x01\x12<\n\x0erevision_paths\x18\x02 \x03(\x0b\x32$.gitaly.GetBlobsRequest.RevisionPath\x12\r\n\x05limit\x18\x03 \x01(\x03\x1a.\n\x0cRevisionPath\x12\x10\n\x08revision\x18\x01 \x01(\t\x12\x0c\n\x04path\x18\x02 \x01(\x0c\"\xa1\x01\n\x10GetBlobsResponse\x12\x0c\n\x04size\x18\x01 \x01(\x03\x12\x0c\n\x04\x64\x61ta\x18\x02 \x01(\x0c\x12\x0b\n\x03oid\x18\x03 \x01(\t\x12\x14\n\x0cis_submodule\x18\x04 \x01(\x08\x12\x0c\n\x04mode\x18\x05 \x01(\x05\x12\x10\n\x08revision\x18\x06 \x01(\t\x12\x0c\n\x04path\x18\x07 \x01(\x0c\x12 \n\x04type\x18\x08 \x01(\x0e\x32\x12.gitaly.ObjectType\"\x8b\x01\n\x10ListBlobsRequest\x12,\n\nrepository\x18\x01 \x01(\x0b\x32\x12.gitaly.RepositoryB\x04\x98\xc6,\x01\x12\x11\n\trevisions\x18\x02 \x03(\t\x12\r\n\x05limit\x18\x03 \x01(\r\x12\x13\n\x0b\x62ytes_limit\x18\x04 \x01(\x03\x12\x12\n\nwith_paths\x18\x05 \x01(\x08\"\x81\x01\n\x11ListBlobsResponse\x12-\n\x05\x62lobs\x18\x01 \x03(\x0b\x32\x1e.gitaly.ListBlobsResponse.Blob\x1a=\n\x04\x42lob\x12\x0b\n\x03oid\x18\x01 \x01(\t\x12\x0c\n\x04size\x18\x02 \x01(\x03\x12\x0c\n\x04\x64\x61ta\x18\x03 \x01(\x0c\x12\x0c\n\x04path\x18\x04 \x01(\x0c\"g\n\x13ListAllBlobsRequest\x12,\n\nrepository\x18\x01 \x01(\x0b\x32\x12.gitaly.RepositoryB\x04\x98\xc6,\x01\x12\r\n\x05limit\x18\x02 \x01(\r\x12\x13\n\x0b\x62ytes_limit\x18\x03 \x01(\x03\"y\n\x14ListAllBlobsResponse\x12\x30\n\x05\x62lobs\x18\x01 \x03(\x0b\x32!.gitaly.ListAllBlobsResponse.Blob\x1a/\n\x04\x42lob\x12\x0b\n\x03oid\x18\x01 \x01(\t\x12\x0c\n\x04size\x18\x02 \x01(\x03\x12\x0c\n\x04\x64\x61ta\x18\x03 \x01(\x0c\"Z\n\nLFSPointer\x12\x0c\n\x04size\x18\x01 \x01(\x03\x12\x0c\n\x04\x64\x61ta\x18\x02 \x01(\x0c\x12\x0b\n\x03oid\x18\x03 \x01(\t\x12\x11\n\tfile_size\x18\x04 \x01(\x03\x12\x10\n\x08\x66ile_oid\x18\x05 \x01(\x0c\"W\n\x15GetLFSPointersRequest\x12,\n\nrepository\x18\x01 \x01(\x0b\x32\x12.gitaly.RepositoryB\x04\x98\xc6,\x01\x12\x10\n\x08\x62lob_ids\x18\x02 \x03(\t\"B\n\x16GetLFSPointersResponse\x12(\n\x0clfs_pointers\x18\x01 \x03(\x0b\x32\x12.gitaly.LFSPointer\"h\n\x16ListLFSPointersRequest\x12,\n\nrepository\x18\x01 \x01(\x0b\x32\x12.gitaly.RepositoryB\x04\x98\xc6,\x01\x12\x11\n\trevisions\x18\x02 \x03(\t\x12\r\n\x05limit\x18\x03 \x01(\x05\"C\n\x17ListLFSPointersResponse\x12(\n\x0clfs_pointers\x18\x01 \x03(\x0b\x32\x12.gitaly.LFSPointer\"X\n\x19ListAllLFSPointersRequest\x12,\n\nrepository\x18\x01 \x01(\x0b\x32\x12.gitaly.RepositoryB\x04\x98\xc6,\x01\x12\r\n\x05limit\x18\x03 \x01(\x05\"F\n\x1aListAllLFSPointersResponse\x12(\n\x0clfs_pointers\x18\x01 \x03(\x0b\x32\x12.gitaly.LFSPointer2\xdd\x04\n\x0b\x42lobService\x12\x44\n\x07GetBlob\x12\x16.gitaly.GetBlobRequest\x1a\x17.gitaly.GetBlobResponse\"\x06\xfa\x97(\x02\x08\x02\x30\x01\x12G\n\x08GetBlobs\x12\x17.gitaly.GetBlobsRequest\x1a\x18.gitaly.GetBlobsResponse\"\x06\xfa\x97(\x02\x08\x02\x30\x01\x12J\n\tListBlobs\x12\x18.gitaly.ListBlobsRequest\x1a\x19.gitaly.ListBlobsResponse\"\x06\xfa\x97(\x02\x08\x02\x30\x01\x12S\n\x0cListAllBlobs\x12\x1b.gitaly.ListAllBlobsRequest\x1a\x1c.gitaly.ListAllBlobsResponse\"\x06\xfa\x97(\x02\x08\x02\x30\x01\x12Y\n\x0eGetLFSPointers\x12\x1d.gitaly.GetLFSPointersRequest\x1a\x1e.gitaly.GetLFSPointersResponse\"\x06\xfa\x97(\x02\x08\x02\x30\x01\x12\\\n\x0fListLFSPointers\x12\x1e.gitaly.ListLFSPointersRequest\x1a\x1f.gitaly.ListLFSPointersResponse\"\x06\xfa\x97(\x02\x08\x02\x30\x01\x12\x65\n\x12ListAllLFSPointers\x12!.gitaly.ListAllLFSPointersRequest\x1a\".gitaly.ListAllLFSPointersResponse\"\x06\xfa\x97(\x02\x08\x02\x30\x01\x42\x34Z2gitlab.com/gitlab-org/gitaly/v16/proto/go/gitalypbb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'blob_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z2gitlab.com/gitlab-org/gitaly/v16/proto/go/gitalypb'
  _GETBLOBREQUEST.fields_by_name['repository']._options = None
  _GETBLOBREQUEST.fields_by_name['repository']._serialized_options = b'\230\306,\001'
  _GETBLOBSREQUEST.fields_by_name['repository']._options = None
  _GETBLOBSREQUEST.fields_by_name['repository']._serialized_options = b'\230\306,\001'
  _LISTBLOBSREQUEST.fields_by_name['repository']._options = None
  _LISTBLOBSREQUEST.fields_by_name['repository']._serialized_options = b'\230\306,\001'
  _LISTALLBLOBSREQUEST.fields_by_name['repository']._options = None
  _LISTALLBLOBSREQUEST.fields_by_name['repository']._serialized_options = b'\230\306,\001'
  _GETLFSPOINTERSREQUEST.fields_by_name['repository']._options = None
  _GETLFSPOINTERSREQUEST.fields_by_name['repository']._serialized_options = b'\230\306,\001'
  _LISTLFSPOINTERSREQUEST.fields_by_name['repository']._options = None
  _LISTLFSPOINTERSREQUEST.fields_by_name['repository']._serialized_options = b'\230\306,\001'
  _LISTALLLFSPOINTERSREQUEST.fields_by_name['repository']._options = None
  _LISTALLLFSPOINTERSREQUEST.fields_by_name['repository']._serialized_options = b'\230\306,\001'
  _BLOBSERVICE.methods_by_name['GetBlob']._options = None
  _BLOBSERVICE.methods_by_name['GetBlob']._serialized_options = b'\372\227(\002\010\002'
  _BLOBSERVICE.methods_by_name['GetBlobs']._options = None
  _BLOBSERVICE.methods_by_name['GetBlobs']._serialized_options = b'\372\227(\002\010\002'
  _BLOBSERVICE.methods_by_name['ListBlobs']._options = None
  _BLOBSERVICE.methods_by_name['ListBlobs']._serialized_options = b'\372\227(\002\010\002'
  _BLOBSERVICE.methods_by_name['ListAllBlobs']._options = None
  _BLOBSERVICE.methods_by_name['ListAllBlobs']._serialized_options = b'\372\227(\002\010\002'
  _BLOBSERVICE.methods_by_name['GetLFSPointers']._options = None
  _BLOBSERVICE.methods_by_name['GetLFSPointers']._serialized_options = b'\372\227(\002\010\002'
  _BLOBSERVICE.methods_by_name['ListLFSPointers']._options = None
  _BLOBSERVICE.methods_by_name['ListLFSPointers']._serialized_options = b'\372\227(\002\010\002'
  _BLOBSERVICE.methods_by_name['ListAllLFSPointers']._options = None
  _BLOBSERVICE.methods_by_name['ListAllLFSPointers']._serialized_options = b'\372\227(\002\010\002'
  _globals['_GETBLOBREQUEST']._serialized_start=48
  _globals['_GETBLOBREQUEST']._serialized_end=138
  _globals['_GETBLOBRESPONSE']._serialized_start=140
  _globals['_GETBLOBRESPONSE']._serialized_end=198
  _globals['_GETBLOBSREQUEST']._serialized_start=201
  _globals['_GETBLOBSREQUEST']._serialized_end=389
  _globals['_GETBLOBSREQUEST_REVISIONPATH']._serialized_start=343
  _globals['_GETBLOBSREQUEST_REVISIONPATH']._serialized_end=389
  _globals['_GETBLOBSRESPONSE']._serialized_start=392
  _globals['_GETBLOBSRESPONSE']._serialized_end=553
  _globals['_LISTBLOBSREQUEST']._serialized_start=556
  _globals['_LISTBLOBSREQUEST']._serialized_end=695
  _globals['_LISTBLOBSRESPONSE']._serialized_start=698
  _globals['_LISTBLOBSRESPONSE']._serialized_end=827
  _globals['_LISTBLOBSRESPONSE_BLOB']._serialized_start=766
  _globals['_LISTBLOBSRESPONSE_BLOB']._serialized_end=827
  _globals['_LISTALLBLOBSREQUEST']._serialized_start=829
  _globals['_LISTALLBLOBSREQUEST']._serialized_end=932
  _globals['_LISTALLBLOBSRESPONSE']._serialized_start=934
  _globals['_LISTALLBLOBSRESPONSE']._serialized_end=1055
  _globals['_LISTALLBLOBSRESPONSE_BLOB']._serialized_start=766
  _globals['_LISTALLBLOBSRESPONSE_BLOB']._serialized_end=813
  _globals['_LFSPOINTER']._serialized_start=1057
  _globals['_LFSPOINTER']._serialized_end=1147
  _globals['_GETLFSPOINTERSREQUEST']._serialized_start=1149
  _globals['_GETLFSPOINTERSREQUEST']._serialized_end=1236
  _globals['_GETLFSPOINTERSRESPONSE']._serialized_start=1238
  _globals['_GETLFSPOINTERSRESPONSE']._serialized_end=1304
  _globals['_LISTLFSPOINTERSREQUEST']._serialized_start=1306
  _globals['_LISTLFSPOINTERSREQUEST']._serialized_end=1410
  _globals['_LISTLFSPOINTERSRESPONSE']._serialized_start=1412
  _globals['_LISTLFSPOINTERSRESPONSE']._serialized_end=1479
  _globals['_LISTALLLFSPOINTERSREQUEST']._serialized_start=1481
  _globals['_LISTALLLFSPOINTERSREQUEST']._serialized_end=1569
  _globals['_LISTALLLFSPOINTERSRESPONSE']._serialized_start=1571
  _globals['_LISTALLLFSPOINTERSRESPONSE']._serialized_end=1641
  _globals['_BLOBSERVICE']._serialized_start=1644
  _globals['_BLOBSERVICE']._serialized_end=2249
# @@protoc_insertion_point(module_scope)
